import pandas as pd
from typing import Optional


def generate_dollar_bar_dataframe(
    tick_data: pd.DataFrame,
    dollar_per_bar: int = 100000,
    number_bars: Optional[int] = None
) -> pd.DataFrame:
    """
    Generates a dollar bar dataframe.

    :param tick_data: DataFrame of tick data.
    :type tick_data: pd.DataFrame
    :param dollar_per_bar: Dollars in each bar, defaults to 100000.
    :type dollar_per_bar: int, optional
    :param number_bars: Number of bars, defaults to None.
    :type number_bars: Optional[int], optional
    :return: A dataframe containing OHLCV data and other relevant information based on dollar bars.
    :rtype: pd.DataFrame
    """

    tick_data['dollar'] = tick_data['price'] * tick_data['size']
    tick_data['dollars_cumulated'] = tick_data['dollar'].cumsum()

    if not dollar_per_bar:
        dollars_total = tick_data['dollars_cumulated'].values[-1]
        dollar_per_bar = dollars_total // number_bars
        dollar_per_bar = round(dollar_per_bar, -2)

    tick_grouped = tick_data.reset_index().assign(grouping_id=lambda x: x.dollars_cumulated // dollar_per_bar)
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')
    ohlcv_dataframe = create_ohlcv_dataframe(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe


def generate_tick_bar_dataframe(
    tick_data: pd.DataFrame,
    tick_per_bar: int = 10,
    number_bars: Optional[int] = None
) -> pd.DataFrame:
    """
    Generates a tick bar dataframe.

    :param tick_data: DataFrame of tick data.
    :type tick_data: pd.DataFrame
    :param tick_per_bar: Number of ticks in each bar, defaults to 10.
    :type tick_per_bar: int, optional
    :param number_bars: Number of bars, defaults to None.
    :type number_bars: Optional[int], optional
    :return: A dataframe containing OHLCV data and other relevant information based on tick bars.
    :rtype: pd.DataFrame
    """

    if not tick_per_bar:
        tick_per_bar = tick_data.shape[0] // number_bars

    tick_grouped = tick_data.reset_index().assign(grouping_id=lambda x: x.index // tick_per_bar)
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')
    ohlcv_dataframe = create_ohlcv_dataframe(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe


def generate_time_bar_dataframe(
    tick_data: pd.DataFrame,
    frequency: str = "5Min"
) -> pd.DataFrame:
    """
    Generates a time bar dataframe.

    :param tick_data: DataFrame of tick data.
    :type tick_data: pd.DataFrame
    :param frequency: Frequency for rounding date time, defaults to "5Min".
    :type frequency: str, optional
    :return: A dataframe containing OHLCV data and other relevant information based on time bars with the specified frequency.
    :rtype: pd.DataFrame
    """

    tick_data_grouped = tick_data.groupby(pd.Grouper(freq=frequency))
    ohlcv_dataframe = create_ohlcv_dataframe(tick_data_grouped)

    return ohlcv_dataframe


def generate_volume_bar_dataframe(
    tick_data: pd.DataFrame,
    volume_per_bar: int = 10000,
    number_bars: Optional[int] = None
) -> pd.DataFrame:
    """
    Generates a volume bar dataframe.

    :param tick_data: DataFrame of tick data.
    :type tick_data: pd.DataFrame
    :param volume_per_bar: Volumes in each bar, defaults to 10000.
    :type volume_per_bar: int, optional
    :param number_bars: Number of bars, defaults to None.
    :type number_bars: Optional[int], optional
    :return: A dataframe containing OHLCV data and other relevant information based on volume bars.
    :rtype: pd.DataFrame
    """

    tick_data['volume_cumulated'] = tick_data['size'].cumsum()

    if not volume_per_bar:
        volume_total = tick_data['volume_cumulated'].values[-1]
        volume_per_bar = volume_total // number_bars
        volume_per_bar = round(volume_per_bar, -2)

    tick_grouped = tick_data.reset_index().assign(grouping_id=lambda x: x.volume_cumulated // volume_per_bar)
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')
    ohlcv_dataframe = create_ohlcv_dataframe(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe




import numpy as np
from numba import jit
from joblib import Parallel, delayed
from typing import Tuple, Callable
from itertools import combinations
from joblib import Parallel, delayed


@jit(nopython=True)
def sharpe_ratio(
    returns: np.ndarray, 
    risk_free_rate: float = 0.0
) -> float:
    """
    Calculate the Sharpe Ratio for a given set of returns.

    :param returns: An array of returns for a portfolio.
    :param risk_free_rate: The risk-free rate.
    :return: The calculated Sharpe Ratio.

    .. math::

        \text{Sharpe Ratio} = \frac{\text{Mean Portfolio Return} - \text{Risk-Free Rate}}
                                  {\text{Standard Deviation of Portfolio Returns}}
    """
    excess_returns = returns - risk_free_rate
    std = np.std(excess_returns)
    if std != 0:
        
        return np.mean(excess_returns) / std
    
    else:
        return 0

def performance_evaluation(
    train_partition: np.ndarray,
    test_partition: np.ndarray,
    n_strategies: int,
    metric: Callable,
    risk_free_return: float
) -> Tuple[bool, float]:
    """
    Evaluate the performance of various strategies on given train and test partitions and 
    compute the logit value to determine if the best in-sample strategy is overfitting.

    :param train_partition: Training data partition used for evaluating in-sample performance.
    :type train_partition: np.ndarray
    :param test_partition: Testing data partition used for evaluating out-of-sample performance.
    :type test_partition: np.ndarray
    :param n_strategies: Number of strategies to evaluate.
    :type n_strategies: int
    :param metric: Metric function for evaluating strategy performance. 
                   The function should accept a data array and risk_free_return as arguments.
    :type metric: Callable
    :param risk_free_return: Risk-free return used in the metric function, often used for Sharpe ratio.
    :type risk_free_return: float

    :return: Tuple where the first value indicates if the best in-sample strategy is overfitting 
             (True if overfitting, False otherwise) and the second value is the logit value computed.
    :rtype: Tuple[bool, float]
    """

    evaluate_train = list(map(lambda i: metric(train_partition[:, i], risk_free_return), range(n_strategies)))
    best_strategy = np.argmax(evaluate_train)
    
    evaluate_test = list(map(lambda i: metric(test_partition[:, i], risk_free_return), range(n_strategies)))
    
    rank_of_best_is_strategy = np.argsort(evaluate_test).tolist().index(best_strategy) + 1
    
    w_bar = rank_of_best_is_strategy / (n_strategies + 1)
    logit_value = np.log(w_bar / (1 - w_bar))
    
    return logit_value <= 0.0, logit_value

def probability_of_backtest_overfitting(
    performances: np.ndarray, 
    n_partitions: int = 16, 
    risk_free_return: float = 0.0,
    metric: Callable = None, 
    n_jobs: int = 1
) -> Tuple[float, np.ndarray]:
    """
    Computes the Probability Of Backtest Overfitting.

    For instance, if \(S=16\), we will form 12,780 combinations.

    .. math::
        \left(\begin{array}{c}
        S \\
        S / 2
        \end{array}\right) = \prod_{i=0}^{S / 2^{-1}} \frac{S-i}{S / 2-i}

    :param performances: Matrix of T×N for T observations on N strategies.
    :type performances: np.ndarray
    :param n_partitions: Number of partitions (must be even).
    :type n_partitions: int
    :param metric: Metric function for evaluating strategy.
    :type metric: Callable
    :param risk_free_return: Risk-free return for calculating Sharpe ratio.
    :type risk_free_return: float
    :param n_jobs: Number of parallel jobs.
    :type n_jobs: int

    :return: Tuple containing Probability Of Backtest Overfitting and an array of logit values.
    :rtype: Tuple[float, List[float]]
    """

    if n_partitions % 2 == 1:
        raise ValueError("Number of partitions must be even.")
    
    if metric is None:
        metric = sharpe_ratio
    
    _, n_strategies = performances.shape
    partitions = np.array_split(performances, n_partitions)
    partition_indices = range(n_partitions)
    partition_combinations_indices = list(combinations(partition_indices, n_partitions // 2))

    results = Parallel(n_jobs=n_jobs)(
        delayed(performance_evaluation)(
            np.concatenate([partitions[i] for i in train_indices], axis=0),
            np.concatenate([partitions[i] for i in partition_indices 
                            if i not in train_indices], axis=0),
            n_strategies, 
            metric, 
            risk_free_return
        ) 
        for train_indices in partition_combinations_indices
    )

    results = np.array(results)  

    pbo = results[:, 0].mean(axis=0)
    logit_values = results[:, 1]
    
    return pbo, logit_values



import numpy as np
from random import gauss
from itertools import product


def synthetic_back_testing(
    forecast: float,
    half_life: float,
    sigma: float,
    n_iteration: int = 100000,
    maximum_holding_period: int = 100,
    profit_taking_range: np.ndarray = np.linspace(0.5, 10, 20),
    stop_loss_range: np.ndarray = np.linspace(0.5, 10, 20),
    seed: int = 0
) -> list[tuple[float, float, float, float, float]]:
    """
    Perform backtesting on synthetic price data generated using the Ornstein-Uhlenbeck process.

    The Ornstein-Uhlenbeck process is given by:
    .. math:: P_t = (1 - \\rho) * F + \\rho * P_{t-1} + \\sigma * Z_t

    where:
    - \(P_t\) is the price at time t
    - \(F\) is the forecast price
    - \(\\rho\) is the autoregression coefficient
    - \(\\sigma\) is the standard deviation of noise
    - \(Z_t\) is a random noise with mean 0 and standard deviation 1

    Args:
        forecast (float): The forecasted price.
        half_life (float): The half-life time needed to reach half.
        sigma (float): The standard deviation of the noise.
        n_iteration (int): Number of iterations. Defaults to 100000.
        maximum_holding_period (int): Maximum holding period. Defaults to 100.
        profit_taking_range (np.ndarray): Profit taking range. Defaults to np.linspace(0.5, 10, 20).
        stop_loss_range (np.ndarray): Stop loss range. Defaults to np.linspace(0.5, 10, 20).
        seed (int): Initial seed value. Defaults to 0.

    Returns:
        list[tuple[float, float, float, float, float]]: List of tuples containing profit taking, stop loss, mean,
        standard deviation, and Sharpe ratio.
    """
    rho = 2 ** (-1. / half_life)  # compute ρ coefficient from half-life
    back_test = []

    for profit_taking, stop_loss in product(profit_taking_range, stop_loss_range):
        stop_prices = []
        for iteration in range(n_iteration):
            price, holding_period = seed, 0  # initial price
            while True:
                price = (1 - rho) * forecast + rho * price + sigma * gauss(0, 1)  # update price using O_U process
                gain = price - seed  # compute gain
                holding_period += 1
                if gain > profit_taking or gain < -stop_loss or holding_period > maximum_holding_period:  # check stop condition
                    stop_prices.append(gain)
                    break
        mean, std = np.mean(stop_prices), np.std(stop_prices)  # compute mean and std of samples
        back_test.append((profit_taking, stop_loss, mean, std, mean / std))  # add mean, std, and Sharpe ratio to backTest data

    return back_test



from collections import Counter
from math import log2
from typing import Dict, Tuple


def plug_in_entropy_estimator(
    message: str,
    approximate_word_length: int = 1
) -> float:
    """
    Calculate Plug-in Entropy Estimator.

    :param message: Input encoded message
    :type message: str
    :param approximate_word_length: Approximation of word length, default is 1
    :type approximate_word_length: int
    :return: Calculated Plug-in Entropy Estimator
    :rtype: float
    """
    pmf = probability_mass_function(message, approximate_word_length)
    plug_in_entropy_estimator = -sum(
        pmf[key] * log2(pmf[key])
        for key in pmf.keys()
    ) / approximate_word_length

    return plug_in_entropy_estimator




import numpy as np
import pandas as pd
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.ensemble import BaggingClassifier
from sklearn.pipeline import Pipeline

from RiskLabAI.backtest.validation.cross_validator_controller import CrossValidatorController


class MyPipeline(Pipeline):
    """
    Custom pipeline class to include sample_weight in fit_params.
    """

    def fit(
        self, 
        X: pd.DataFrame, 
        y: pd.DataFrame, 
        sample_weight: list = None, 
        **fit_params
    ) -> 'MyPipeline':
        """
        Fit the pipeline while considering sample weights.
        
        :param X: Feature data.
        :param y: Labels of data.
        :param sample_weight: Sample weights for fit, defaults to None.
        :param **fit_params: Additional fit parameters.
        :return: Fitted pipeline.
        """
        if sample_weight is not None:
            fit_params[self.steps[-1][0] + '__sample_weight'] = sample_weight
        return super(MyPipeline, self).fit(X, y, **fit_params)


def clf_hyper_fit(
    feature_data: pd.DataFrame,
    label: pd.DataFrame,
    times: pd.Series,
    pipe_clf: Pipeline,
    param_grid: dict,
    validator_type: str = 'purgedkfold',
    validator_params: dict = None,
    bagging: list = [0, -1, 1.],
    rnd_search_iter: int = 0,
    n_jobs: int = -1,
    **fit_params
) -> MyPipeline:
    """
    Perform hyperparameter tuning and model fitting.

    :param feature_data: Data of features.
    :param label: Labels of data.
    :param times: The timestamp series associated with the labels.
    :param pipe_clf: Our estimator.
    :param param_grid: Parameter space.
    :param validator_type: Type of cross-validator to create.
    :param validator_params: Additional keyword arguments to be passed to the cross-validator's constructor.
    :param bagging: Bagging type, defaults to [0, -1, 1.].
    :param rnd_search_iter: Number of iterations for randomized search, defaults to 0.
    :param n_jobs: Number of jobs for parallel processing, defaults to -1.
    :param **fit_params: Additional fit parameters.
    :return: Fitted pipeline.
    """
    if set(label.values) == {0, 1}:
        scoring = 'f1'  # F1-score for meta-labeling
    else:
        scoring = 'neg_log_loss'  # Symmetric towards all cases

    if validator_params is None:
        validator_params = {
            'times' : times,
            'n_splits' : 5,
            'embargo' : 0.01,
        }

    # Hyperparameter search on train data
    inner_cv =  CrossValidatorController(
        validator_type,
        **validator_params
    ).cross_validator
    
    if rnd_search_iter == 0:
        gs = GridSearchCV(estimator=pipe_clf, param_grid=param_grid, scoring=scoring, cv=inner_cv, n_jobs=n_jobs)
    else:
        gs = RandomizedSearchCV(estimator=pipe_clf, param_distributions=param_grid, scoring=scoring,
                                cv=inner_cv, n_jobs=n_jobs, n_iter=rnd_search_iter)
    gs = gs.fit(feature_data, label, **fit_params).best_estimator_  # Pipeline

    # Fit validated model on the entirety of the data
    if bagging[1] > 0:
        gs = BaggingClassifier(base_estimator=MyPipeline(gs.steps), n_estimators=int(bagging[0]),
                               max_samples=float(bagging[1]), max_features=float(bagging[2]), n_jobs=n_jobs)
        gs = gs.fit(feature_data, label, sample_weight=fit_params[gs.base_estimator.steps[-1][0] + '__sample_weight'])
        gs = Pipeline([('bag', gs)])

    return gs



from math import pi
import pandas as pd
from RiskLabAI.features.microstructural_features.corwin_schultz import beta_estimates, gamma_estimates


def sigma_estimates(
    beta: pd.Series,
    gamma: pd.Series
) -> pd.Series:
    """
    Compute Bekker-Parkinson volatility σ estimates.

    This function calculates the Bekker-Parkinson volatility estimates based on the provided
    beta and gamma values. The mathematical formula used is:

    .. math::
        \sigma = \frac{(2^{0.5} - 1) \cdot (\beta ^ {0.5})}{3 - 2 \cdot (2^{0.5})}
                + \left(\frac{\gamma}{\left(\frac{8}{\pi}\right)^{0.5} \cdot (3 - 2 \cdot (2^{0.5}))}\right)^{0.5}

    Negative resulting values are set to 0.

    :param beta: β Estimates vector.
    :param gamma: γ Estimates vector.
    :return: Bekker-Parkinson volatility σ estimates.

    Reference:
        De Prado, M. (2018) Advances in Financial Machine Learning, page 286, snippet 19.2.
    """
    k2 = (8 / pi) ** 0.5
    denominator = 3 - 2 * (2 ** 0.5)

    sigma = (2 ** 0.5 - 1) * (beta ** 0.5) / denominator
    sigma += (gamma / (k2 ** 2 * denominator)) ** 0.5
    sigma[sigma < 0] = 0

    return sigma


def bekker_parkinson_volatility_estimates(
    high_prices: pd.Series,
    low_prices: pd.Series,
    window_span: int = 20
) -> pd.Series:
    """
    Compute Bekker-Parkinson volatility estimates based on high and low prices.

    Utilizes Corwin and Schultz estimation techniques to calculate the Bekker-Parkinson
    volatility. The function first determines the beta and gamma values and then
    uses them to compute the volatility estimates.

    :param high_prices: High prices vector.
    :param low_prices: Low prices vector.
    :param window_span: Rolling window span for beta estimation.
    :return: Bekker-Parkinson volatility estimates.

    Reference:
        De Prado, M. (2018) Advances in Financial Machine Learning, page 286, "Corwin and Schultz" section.
    """
    beta = beta_estimates(high_prices, low_prices, window_span)
    gamma = gamma_estimates(high_prices, low_prices)

    return sigma_estimates(beta, gamma)



import sys
import time
from typing import NoReturn


def progress_bar(
    current_progress: int, 
    total_progress: int, 
    start_time: float, 
    bar_length: int = 20
) -> NoReturn:
    """
    Display a terminal-style progress bar with completion percentage and estimated remaining time.

    :param current_progress: Current value indicating the progress made.
    :param total_progress: Total value representing the completion of the task.
    :param start_time: The time at which the task started, typically acquired via time.time().
    :param bar_length: Length of the progress bar in terminal characters, default is 20.
    
    The displayed progress bar uses the formula:

    .. math::
        \text{percentage} = \frac{\text{current\_progress}}{\text{total\_progress}}

    The estimated remaining time is calculated based on elapsed time and progress made:

    .. math::
        \text{remaining\_time} = \frac{\text{elapsed\_time} \times (\text{total\_progress} - \text{current\_progress})}{\text{current\_progress}}

    :return: None
    """

    percentage = current_progress / total_progress
    arrow = '-' * int(round(percentage * bar_length) - 1) + '>'
    spaces = ' ' * (bar_length - len(arrow))

    elapsed_time = time.time() - start_time
    
    if current_progress == 0:
        remaining_time = "Calculating..."
    else:
        remaining_time = int(
            (elapsed_time / current_progress) * (total_progress - current_progress) / 60
        )

    if current_progress == total_progress:
        sys.stdout.write("\rCompleted: [{0}] 100% - Task completed!\n".format('-' * bar_length))
    
    else:
        sys.stdout.write("\rCompleted: [{0}] {1}% - {2} minutes remaining.".format(
            arrow + spaces, int(round(percentage * 100)), remaining_time))
    
    sys.stdout.flush()



import sys
import time
import numpy as np
import pandas as pd
from typing import Tuple
from RiskLabAI.data.structures.utilities_lopez import *

def progress_bar(
    value: int,
    end_value: int,
    start_time: float,
    bar_length: int = 20
) -> None:
    """
    Display a progress bar in the console.

    :param value: Current progress value.
    :param end_value: The end value indicating 100% progress.
    :param start_time: Time when the event started.
    :param bar_length: The length of the progress bar in characters. Default is 20.
    """
    percent = float(value) / end_value
    arrow = '-' * int(round(percent * bar_length) - 1) + '>'
    spaces = ' ' * (bar_length - len(arrow))
    remaining = int(((time.time() - start_time) / value) * (end_value - value) / 60)

    sys.stdout.write(
        "\rCompleted: [{0}] {1}% - {2} minutes remaining.".format(
            arrow + spaces, int(round(percent * 100)), remaining
        )
    )
    sys.stdout.flush()


def ewma(
    input_array: np.ndarray,
    window_length: int
) -> np.ndarray:
    """
    Computes the Exponentially Weighted Moving Average (EWMA).

    :param input_array: The input time series array.
    :param window_length: Window length for the EWMA.
    :return: The EWMA values.
    """
    N = input_array.shape[0]
    output_ewma = np.empty(N, dtype='float64')
    omega = 1
    alpha = 2 / float(window_length + 1)
    current_value = input_array[0]
    output_ewma[0] = current_value

    for i in range(1, N):
        omega += (1 - alpha) ** i
        current_value = current_value * (1 - alpha) + input_array[i]
        output_ewma[i] = current_value / omega

    return output_ewma


def compute_grouping(
    target_col: pd.Series,
    initial_expected_ticks: int,
    bar_size: float
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """
    Group a DataFrame based on a feature and calculates thresholds.

    :param target_col: Target column of tick dataframe.
    :param initial_expected_ticks: Initial expected ticks.
    :param bar_size: Initial expected size in each tick.
    :return: Arrays of times_delta, thetas_absolute, thresholds, times, thetas, grouping_id.
    """
    N = target_col.shape[0]
    target_col = target_col.values.astype(np.float64)
    thetas_absolute, thresholds, thetas, grouping_id = np.zeros(N), np.zeros(N), np.zeros(N), np.zeros(N)
    thetas_absolute[0], current_theta = np.abs(target_col[0]), target_col[0]
    start_time = time.time()
    current_group_id, time_prev, expected_ticks, expected_bar_value = 0, 0, initial_expected_ticks, bar_size
    times_delta, times = [], []

    for i in range(1, N):
        current_theta += target_col[i]
        thetas[i] = current_theta
        theta_absolute = np.abs(current_theta)
        thetas_absolute[i] = theta_absolute
        threshold = expected_ticks * expected_bar_value
        thresholds[i] = threshold
        grouping_id[i] = current_group_id

        if theta_absolute >= threshold:
            current_group_id += 1
            current_theta = 0
            times_delta.append(np.float64(i - time_prev))
            times.append(i)
            time_prev = i
            expected_ticks = ewma(np.array(times_delta), window_length=np.int64(len(times_delta)))[-1]
            expected_bar_value = np.abs(ewma(target_col[:i], window_length=np.int64(initial_expected_ticks))[0])

        progress_bar(i, N, start_time)

    return times_delta, thetas_absolute, thresholds, times, thetas, grouping_id

import pandas as pd
import numpy as np
from typing import Tuple


def generate_information_driven_bars(
    tick_data: pd.DataFrame,
    bar_type: str = "volume",
    tick_expected_initial: int = 2000
) -> Tuple[pd.DataFrame, np.array, np.array]:
    """
    Implements Information-Driven Bars as per the methodology described in
    "Advances in financial machine learning" by De Prado (2018).

    :param tick_data: DataFrame of tick data.
    :param bar_type: Type of the bars, options: "tick", "volume", "dollar".
    :param tick_expected_initial: Initial expected ticks value.
    :return: A tuple containing the OHLCV DataFrame, thetas absolute array,
        and thresholds array.
    """
    if bar_type == "volume":
        input_data = tick_data['volume_labeled']
    elif bar_type == "tick":
        input_data = tick_data['label']
    elif bar_type == "dollar":
        input_data = tick_data['dollarslabeled']
    else:
        raise ValueError("Invalid bar_type provided. Choose among 'tick', 'volume', 'dollar'.")

    bar_expected_value = np.abs(input_data.mean())

    times_delta, thetas_absolute, thresholds, _, _, grouping_id = compute_grouping(
        input_data, tick_expected_initial, bar_expected_value)

    tick_grouped = tick_data.reset_index().assign(grouping_id=grouping_id)
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']

    ohlcv_dataframe = ohlcv(tick_grouped.groupby('grouping_id'))
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe, thetas_absolute, thresholds


def ohlcv(
    tick_data_grouped: pd.core.groupby.generic.DataFrameGroupBy
) -> pd.DataFrame:
    """
    Computes various statistics for the grouped tick data.

    Takes a grouped dataframe, combines the data, and creates a new one with
    information about prices, volume, and other statistics. This is typically
    used in the context of financial tick data to generate OHLCV data
    (Open, High, Low, Close, Volume).

    :param tick_data_grouped: Grouped DataFrame containing tick data.
    :return: A DataFrame containing OHLCV data and other derived statistics.
    """
    ohlc = tick_data_grouped['price'].ohlc()
    ohlc['volume'] = tick_data_grouped['size'].sum()
    ohlc['value_of_trades'] = tick_data_grouped.apply(
        lambda x: (x['price'] * x['size']).sum() / x['size'].sum()
    )
    ohlc['price_mean'] = tick_data_grouped['price'].mean()
    ohlc['tick_count'] = tick_data_grouped['price'].count()
    ohlc['price_mean_log_return'] = np.log(ohlc['price_mean']) - np.log(ohlc['price_mean'].shift(1))

    return ohlc


def generate_time_bar(
    tick_data: pd.DataFrame,
    frequency: str = "5Min"
) -> pd.DataFrame:
    """
    Generates time bars for tick data.

    This function groups tick data by a specified time frequency and then
    computes OHLCV (Open, High, Low, Close, Volume) statistics.

    :param tick_data: DataFrame containing tick data.
    :param frequency: Time frequency for rounding datetime.
    :return: A DataFrame containing OHLCV data grouped by time.
    """
    tick_data_grouped = tick_data.groupby(pd.Grouper(freq=frequency))
    ohlcv_dataframe = ohlcv(tick_data_grouped)
    return ohlcv_dataframe


def generate_tick_bar(
    tick_data: pd.DataFrame,
    ticks_per_bar: int = 10,
    number_bars: int = None
) -> pd.DataFrame:
    """
    Generates tick bars for tick data.

    This function groups tick data by a specified number of ticks and then
    computes OHLCV statistics.

    :param tick_data: DataFrame containing tick data.
    :param ticks_per_bar: Number of ticks in each bar.
    :param number_bars: Number of bars to generate.
    :return: A DataFrame containing OHLCV data grouped by tick count.
    """
    if not ticks_per_bar:
        ticks_per_bar = int(tick_data.shape[0] / number_bars)

    tick_grouped = tick_data.reset_index().assign(grouping_id=lambda x: x.index // ticks_per_bar)
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')
    ohlcv_dataframe = ohlcv(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe

def generate_volume_bar(
    tick_data: pd.DataFrame,
    volume_per_bar: int = 10000,
    number_bars: int = None
) -> pd.DataFrame:
    """
    Generates volume bars for tick data.

    This function groups tick data by a specified volume size and then computes OHLCV statistics.

    :param tick_data: DataFrame containing tick data.
    :param volume_per_bar: Volume size for each bar.
    :param number_bars: Number of bars to generate.

    :return: A DataFrame containing OHLCV data grouped by volume.
    """
    tick_data['volume_cumulated'] = tick_data['size'].cumsum()

    if not volume_per_bar:
        volume_total = tick_data['volume_cumulated'].values[-1]
        volume_per_bar = volume_total / number_bars
        volume_per_bar = round(volume_per_bar, -2)

    tick_grouped = tick_data.reset_index().assign(
        grouping_id=lambda x: x['volume_cumulated'] // volume_per_bar
    )
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')
    ohlcv_dataframe = ohlcv(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe


def generate_dollar_bar(
    tick_data: pd.DataFrame,
    dollar_per_bar: float = 100000,
    number_bars: int = None
) -> pd.DataFrame:
    """
    Generates dollar bars for tick data.

    This function groups tick data by a specified dollar amount and then computes OHLCV statistics.

    :param tick_data: DataFrame containing tick data.
    :param dollar_per_bar: Dollar amount for each bar.
    :param number_bars: Number of bars to generate.

    :return: A DataFrame containing OHLCV data grouped by dollar amount.
    """
    tick_data['dollar'] = tick_data['price'] * tick_data['size']
    tick_data['dollars_cumulated'] = tick_data['dollar'].cumsum()

    if not dollar_per_bar:
        dollars_total = tick_data['dollars_cumulated'].values[-1]
        dollar_per_bar = dollars_total / number_bars
        dollar_per_bar = round(dollar_per_bar, -2)

    tick_grouped = tick_data.reset_index().assign(
        grouping_id=lambda x: x['dollars_cumulated'] // dollar_per_bar
    )
    dates = tick_grouped.groupby('grouping_id', as_index=False).first()['dates']
    tick_data_grouped = tick_grouped.groupby('grouping_id')
    ohlcv_dataframe = ohlcv(tick_data_grouped)
    ohlcv_dataframe.set_index(dates, drop=True, inplace=True)

    return ohlcv_dataframe


def calculate_pca_weights(
    covariance_matrix: np.ndarray,
    risk_distribution: np.ndarray = None,
    risk_target: float = 1.0
) -> np.ndarray:
    """
    Calculates hedging weights using the covariance matrix, risk distribution, and risk target.

    :param covariance_matrix: Covariance matrix.
    :param risk_distribution: Risk distribution vector.
    :param risk_target: Risk target value.

    :return: Weights.
    """
    eigenvalues, eigenvectors = np.linalg.eigh(covariance_matrix)
    indices = eigenvalues.argsort()[::-1]
    eigenvalues, eigenvectors = eigenvalues[indices], eigenvectors[:, indices]

    if risk_distribution is None:
        risk_distribution = np.zeros(covariance_matrix.shape[0])
        risk_distribution[-1] = 1.0

    loads = risk_target * (risk_distribution / eigenvalues) ** 0.5
    weights = np.dot(eigenvectors, np.reshape(loads, (-1, 1)))

    return weights


def events(
    input_data: pd.DataFrame,
    threshold: float
) -> pd.DatetimeIndex:
    """
    Implementation of the symmetric CUSUM filter.

    This function computes time events when certain price change thresholds are met.

    :param input_data: DataFrame of prices and dates.
    :param threshold: Threshold for price change.

    :return: DatetimeIndex containing events.
    """
    time_events, shift_positive, shift_negative = [], 0, 0
    price_delta = input_data.diff()

    for i in price_delta.index[1:]:
        shift_positive = max(0, shift_positive + price_delta.loc[i])
        shift_negative = min(0, shift_negative + price_delta.loc[i])

        if shift_negative < -threshold:
            shift_negative = 0
            time_events.append(i)
        elif shift_positive > threshold:
            shift_positive = 0
            time_events.append(i)

    return pd.DatetimeIndex(time_events)



from .abstract_bars import *
from .abstract_imbalance_bars import *
from .abstract_information_driven_bars import *
from .abstract_run_bars import *
from .data_structures_lopez import *
from .filtering_lopez import *
from .hedging import *
from .imbalance_bars import *
from .infomation_driven_bars import *
from .run_bars import *
from .standard_bars_lopez import *
from .standard_bars import *
from .time_bars import *
from .utilities_lopez import *



import numpy as np
import pandas as pd

def beta_estimates(
    high_prices: pd.Series,
    low_prices: pd.Series,
    window_span: int
) -> pd.Series:
    """
    Estimate β using Corwin and Schultz methodology.

    :param high_prices: High prices vector.
    :param low_prices: Low prices vector.
    :param window_span: Rolling window span.
    :return: Estimated β vector.

    .. note:: Reference: Corwin, S. A., & Schultz, P. (2012). A simple way to estimate bid-ask spreads from daily high and low prices. The Journal of Finance, 67(2), 719-760.
    """
    log_ratios = np.log(high_prices / low_prices) ** 2
    beta = log_ratios.rolling(window=2).sum()
    beta = beta.rolling(window=window_span).mean()
    return beta

def gamma_estimates(
    high_prices: pd.Series,
    low_prices: pd.Series
) -> pd.Series:
    """
    Estimate γ using Corwin and Schultz methodology.

    :param high_prices: High prices vector.
    :param low_prices: Low prices vector.
    :return: Estimated γ vector.

    .. note:: Reference: Corwin, S. A., & Schultz, P. (2012). A simple way to estimate bid-ask spreads from daily high and low prices. The Journal of Finance, 67(2), 719-760.
    """
    high_prices_max = high_prices.rolling(window=2).max()
    low_prices_min = low_prices.rolling(window=2).min()
    gamma = np.log(high_prices_max / low_prices_min)**2
    return gamma

def alpha_estimates(
    beta: pd.Series,
    gamma: pd.Series
) -> pd.Series:
    """
    Estimate α using Corwin and Schultz methodology.

    :param beta: β Estimates vector.
    :param gamma: γ Estimates vector.
    :return: Estimated α vector.

    .. note:: Reference: Corwin, S. A., & Schultz, P. (2012). A simple way to estimate bid-ask spreads from daily high and low prices. The Journal of Finance, 67(2), 719-760.
    """
    denominator = 3 - 2 * 2**0.5
    alpha = (2**0.5 - 1) * (beta**0.5) / denominator
    alpha -= (gamma / denominator)**0.5
    alpha[alpha < 0] = 0
    return alpha

def corwin_schultz_estimator(
    high_prices: pd.Series,
    low_prices: pd.Series,
    window_span: int = 20
) -> pd.Series:
    """
    Estimate spread using Corwin and Schultz methodology.

    :param high_prices: High prices vector.
    :param low_prices: Low prices vector.
    :param window_span: Rolling window span, default is 20.
    :return: Estimated spread vector.

    .. note:: Reference: Corwin, S. A., & Schultz, P. (2012). A simple way to estimate bid-ask spreads from daily high and low prices. The Journal of Finance, 67(2), 719-760.
    """
    beta = beta_estimates(high_prices, low_prices, window_span)
    gamma = gamma_estimates(high_prices, low_prices)
    alpha = alpha_estimates(beta, gamma)
    spread = 2 * (alpha - 1) / (1 + np.exp(alpha))
    return spread



import plotly.graph_objects as go

def update_figure_layout(fig, title, xaxis_title, yaxis_title, legend_x=1, legend_y=1):
    fig.update_layout(
        title=title,
        xaxis_title=xaxis_title,
        yaxis_title=yaxis_title,
        template="plotly_dark",
        plot_bgcolor='rgba(0,0,0,0)',  # Setting transparent background
        paper_bgcolor='rgba(0,0,0,0)', # Setting transparent background
        legend=dict(
            x=legend_x,       # X position of the legend
            y=legend_y,       # Y position of the legend
            xanchor='auto',   # Anchor point of the legend
            yanchor='auto'    # Anchor point of the legend
        )
    )



import csv
from math import floor
from typing import Iterable, Optional, Generator, Union

import pandas as pd

from RiskLabAI.controller.bars_initializer import BarsInitializerController
from RiskLabAI.data.structures.abstract_bars import AbstractBars
from RiskLabAI.utils.constants import (
    DATE_TIME, TICK_NUMBER, OPEN_PRICE, HIGH_PRICE,
    LOW_PRICE, CLOSE_PRICE, CUMULATIVE_VOLUME,
    CUMULATIVE_BUY_VOLUME, CUMULATIVE_SELL_VOLUME,
    CUMULATIVE_TICKS, CUMULATIVE_DOLLAR, THRESHOLD
)


class Controller:
    def __init__(self) -> None:
        self.bars_initializer = BarsInitializerController()

    def handle_input_command(
            self,
            method_name: str,
            method_arguments: dict,
            input_data: Union[str, pd.DataFrame],
            output_path: Optional[str] = None,
            batch_size: int = 1_000_000
    ) -> pd.DataFrame:
        """
        Handles the input command to initialize bars and run on batches.

        :param method_name: Name of the method to call
        :param method_arguments: Arguments for the method
        :param input_data: Input data as a DataFrame or string path
        :param output_path: Optional path to save results as CSV
        :param batch_size: Size of each batch to process
        :return: DataFrame of aggregated bars
        """
        method = self.bars_initializer.method_name_to_method[method_name]
        initialized_bars = method(**method_arguments)

        return self.run_on_batches(
            initialized_bars=initialized_bars,
            input_data=input_data,
            batch_size=batch_size,
            output_path=output_path
        )

    def run_on_batches(
            self,
            initialized_bars: AbstractBars,
            input_data: Union[str, pd.DataFrame],
            batch_size: int,
            output_path: Optional[str] = None
    ) -> pd.DataFrame:
        """
        Runs the initialized bars on batches of data.

        :param initialized_bars: Initialized bars object
        :param input_data: Input data as DataFrame or string path
        :param batch_size: Size of each batch to process
        :param output_path: Optional path to save results as CSV
        :return: DataFrame of aggregated bars
        """
        save_to_csv = output_path is not None

        if save_to_csv:
            with open(output_path, 'w'):
                pass
            add_header = True

        column_names = [
            DATE_TIME, TICK_NUMBER, OPEN_PRICE, HIGH_PRICE,
            LOW_PRICE, CLOSE_PRICE, CUMULATIVE_VOLUME,
            CUMULATIVE_BUY_VOLUME, CUMULATIVE_SELL_VOLUME,
            CUMULATIVE_TICKS, CUMULATIVE_DOLLAR, THRESHOLD
        ]

        aggregated_bars = []

        if isinstance(input_data, str):
            batches_generator = self.read_batches_from_string(
                input_data, batch_size)
        elif isinstance(input_data, pd.DataFrame):
            batches_generator = self.read_batches_from_dataframe(
                input_data, batch_size)
        else:
            raise ValueError('Input data type should be DataFrame or string.')

        for k, batch in enumerate(batches_generator):
            print(f"Processing batch {k} with size {batch.shape[0]}")

            batch_bars = self.construct_bars_from_batch(
                initialized_bars, batch)

            aggregated_bars.extend(batch_bars)

            if save_to_csv:
                pd.DataFrame(
                    batch_bars, columns=column_names
                ).to_csv(output_path, header=add_header, index=False, mode='a')
                add_header = False

        return pd.DataFrame(aggregated_bars, columns=column_names)

    @staticmethod
    def construct_bars_from_batch(
            bars: AbstractBars,
            data: pd.DataFrame
    ) -> list:
        """
        Construct bars from a single batch of data.

        :param bars: Initialized bars object
        :param data: Data for this batch as a DataFrame
        :return: List of constructed bars
        """
        return bars.construct_bars_from_data(data=data.values)

    @staticmethod
    def read_batches_from_string(
            input_path: str,
            batch_size: int
    ) -> Generator:
        """
        Reads data in batches from a CSV file.

        :param input_path: File path to read from
        :param batch_size: Size of each batch
        :return: Generator yielding batches of data
        """
        n_rows = sum(1 for _ in csv.reader(open(input_path)))
        n_batches = max(1, floor(n_rows / batch_size))

        if n_batches == 1:
            yield pd.read_csv(input_path, parse_dates=[0])
        else:
            for batch in pd.read_csv(input_path, chunksize=batch_size, parse_dates=[0]):
                yield batch

    @staticmethod
    def read_batches_from_dataframe(
            input_data: pd.DataFrame,
            batch_size: int
    ) -> Generator:
        """
        Reads data in batches from a DataFrame.

        :param input_data: DataFrame to read from
        :param batch_size: Size of each batch
        :return: Generator yielding batches of data
        """
        n_rows = input_data.shape[0]
        n_batches = max(1, floor(n_rows / batch_size))

        if n_batches == 1:
            yield input_data
        else:
            for k in range(n_batches):
                start_idx = k * batch_size
                end_idx = min((k + 1) * batch_size, n_rows)
                yield input_data.iloc[start_idx:end_idx].copy()



import pandas as pd
import numpy as np
from enum import Enum
from tqdm.notebook import tqdm
from numpy.random import RandomState
from collections import Counter
import itertools
import sklearn.metrics as Metrics
from sklearn.preprocessing import minmax_scale
from pycaret.classification import ClassificationExperiment
from sklearn.model_selection import StratifiedKFold 
from arch.bootstrap import optimal_block_length
from arch.bootstrap import StationaryBootstrap, CircularBlockBootstrap, MovingBlockBootstrap, IndependentSamplesBootstrap, IIDBootstrap
from sklearn.inspection import permutation_importance
from sklearn.feature_selection import mutual_info_classif, f_classif
from scipy.stats import pearsonr, kendalltau


def feature_importance_linear_models(
    classifier,
    X, y,
) -> pd.DataFrame:
    classifier.fit(X, y)
    coefficients_importances = np.abs(classifier.coef_).mean(axis=0)
    coefficients_importances = pd.Series(coefficients_importances, classifier.feature_names_in_)

    importances_scaled = minmax_scale(
        coefficients_importances,
        feature_range=(0, 1),
        axis=0,
    )

    return pd.Series(importances_scaled, index=coefficients_importances.index)


def feature_importance_sklearn(
    classifier,
    X, y,
) -> pd.DataFrame:
    classifier.fit(X, y)
    importances = pd.Series(classifier.feature_importances_, index=classifier.feature_names_in_)
    
    importances = pd.concat({
        "Mean": importances.mean(),
    }, axis=1)

    importances_scaled = minmax_scale(
        importances["Mean"],
        feature_range=(0, 1),
        axis=0,
    )

    return pd.Series(importances_scaled, index=importances.index)


def feature_importance_RFE(
    classifier,
    X,
    y,
) -> pd.DataFrame:
    rfe = RFE(
        estimator=classifier,
        verbose=0,
        n_features_to_select=1,    
    )

    rfe.fit(X, y)

    inverted_ranking = np.max(rfe.ranking_) - rfe.ranking_ + 1

    normalized_importance = minmax_scale(inverted_ranking)

    return pd.Series(normalized_importance, index=rfe.feature_names_in_)

def feature_importance_SFI(
    classifier,
    X,
    y,
    n_splits: int = 5,
    score_sample_weights: list = None,  
    train_sample_weights: list = None, 
) -> pd.DataFrame:

    if train_sample_weights is None:
        train_sample_weights = np.ones(X.shape[0])
    if score_sample_weights is None:
        score_sample_weights = np.ones(X.shape[0])

    cv_generator = StratifiedKFold(n_splits=n_splits)
    feature_names = X.columns
    importances = []
    for feature_name in feature_names:
        scores = []

        for train, test in cv_generator.split(X, y):
            x_train, y_train, weights_train = X.iloc[train, :][[feature_name]], y.iloc[train], train_sample_weights[train]
            x_test, y_test, weights_test = X.iloc[test, :][[feature_name]], y.iloc[test], score_sample_weights[test]

            feature_train, label_train, sample_weights_train = (
                x_train, y_train, weights_train,
            )

            feature_test, label_test, sample_weights_test = (
                x_test, y_test, weights_test,
            )

            try:
                classifier.fit(feature_train, label_train, sample_weight=sample_weights_train)
            except:
                classifier.fit(feature_train, label_train)

            prediction = classifier.predict(feature_test)
            score = Metrics.accuracy_score(label_test, prediction, sample_weight=sample_weights_test)
            scores.append(score)

        importances.append({
            "FeatureName": feature_name,
            "Mean": np.mean(scores),
        })

    importances = pd.DataFrame(importances)
    importances_scaled = minmax_scale(
        importances["Mean"],
        feature_range=(0, 1),
        axis=0,
    )

    return pd.Series(importances_scaled, index=importances.FeatureName)


def feature_importance_MDI(classifier, X, y):
    classifier.fit(X, y)
    importances = pd.Series(classifier.feature_importances_, index=classifier.feature_names_in_)
    
    importances_scaled = minmax_scale(
        importances,
        feature_range=(0, 1),
        axis=0,
    )

    return pd.Series(importances_scaled, index=importances.index)

def feature_importance_MDA(
    classifier,
    X: pd.DataFrame,
    y: pd.DataFrame,
    n_repeats: int = 5,
) -> pd.DataFrame:

    classifier.fit(X, y)
    importances = permutation_importance(classifier, X, y, n_repeats=n_repeats, random_state=43).importances_mean

    importances = minmax_scale(
        importances,
        feature_range=(0, 1),
        axis=0,
    )

    return pd.Series(importances, index=X.columns)

class UAMeasure(Enum):
    SPEARMAN = "Spearman"
    PEARSON = "Pearson"
    KENDAL_TAU = "Kendal-Tau"
    MUTUAL_INFORMATION = "Mutual Information"
    ANOVA_F = "ANOVA F-Stat"

def feature_importance_univariate_analysis(measurements, X, y):
    ua = pd.DataFrame(
        data=None,
        index=X.columns
    )

    if UAMeasure.SPEARMAN.value in measurements:
        spearmans = minmax_scale(X.apply(lambda feature: (np.abs(np.corrcoef(feature, y)[0, 1])), axis=0))
        ua[UAMeasure.SPEARMAN.value] = spearmans

    if UAMeasure.PEARSON.value in measurements:
        pearsons = minmax_scale(X.apply(lambda feature: np.abs(pearsonr(feature, y).statistic), axis=0))
        ua[UAMeasure.PEARSON.value] = pearsons
        
    if UAMeasure.KENDAL_TAU.value in measurements:
        kendalltaus = minmax_scale(X.apply(lambda feature: np.abs(kendalltau(feature, y).statistic), axis=0))
        ua[UAMeasure.KENDAL_TAU.value] = kendalltaus

    if UAMeasure.MUTUAL_INFORMATION.value in measurements:
        mutual_informations = minmax_scale(mutual_info_classif(X, y, n_neighbors=51))
        ua[UAMeasure.MUTUAL_INFORMATION.value] = mutual_informations

    if UAMeasure.ANOVA_F.value in measurements:
        f_values = minmax_scale(f_classif(X, y)[0])
        ua[UAMeasure.ANOVA_F.value] = f_values
    
    return ua

def full_name(klass):
    return ".".join([klass.__module__, klass.__name__])

class FIModel(Enum):
    COEFFICIENT_BASED = "Coefficient-Based + Shrinkage & Selection"
    TREE_BASED = "Tree-Based"
    PERMUTATION_BASED = "Permutation-Based"
    SINGLE_FEATURE_BASED = "Signle Feature-Based"
    RFE_BASED = "Recursive Feature Elimination-Based"

class SWEFI:
    def __init__(self, X, y, n_fold=10):

        self.percentage = None
        self.clfx = ClassificationExperiment()
        self.clfx.setup(data = X, target = y, fold=n_fold, train_size=0.99, session_id = 123, n_jobs = -1, normalize=True, normalize_method='zscore')
        self.X = X
        self.y = y

    def select_models(self, select_n_model=None):
        model_to_methods = self.clfx.models()

        model_to_methods['FI Methods'] = pd.Series({
            'lr': [FIModel.COEFFICIENT_BASED.value, FIModel.PERMUTATION_BASED.value, FIModel.SINGLE_FEATURE_BASED.value, FIModel.RFE_BASED.value],
            'knn': [FIModel.PERMUTATION_BASED.value, ],
            'nb': [FIModel.SINGLE_FEATURE_BASED.value],
            'dt': [FIModel.SINGLE_FEATURE_BASED.value],
            'svm': [FIModel.COEFFICIENT_BASED.value, FIModel.PERMUTATION_BASED.value,], 
            'rbfsvm': [FIModel.PERMUTATION_BASED.value,],
            'gpc': [FIModel.PERMUTATION_BASED.value,],
            'mlp': [FIModel.PERMUTATION_BASED.value,],
            'ridge': [FIModel.COEFFICIENT_BASED.value, FIModel.PERMUTATION_BASED.value, FIModel.SINGLE_FEATURE_BASED.value, FIModel.RFE_BASED.value],
            'rf': [FIModel.TREE_BASED.value, FIModel.RFE_BASED.value],
            'qda': [FIModel.PERMUTATION_BASED.value, ],
            'ada': [FIModel.PERMUTATION_BASED.value, ],
            'gbc': [FIModel.PERMUTATION_BASED.value, ],
            'lda': [FIModel.COEFFICIENT_BASED.value, FIModel.PERMUTATION_BASED.value, FIModel.SINGLE_FEATURE_BASED.value, FIModel.RFE_BASED.value],
            'et': [FIModel.TREE_BASED.value, FIModel.RFE_BASED.value],
            'xgboost': [FIModel.TREE_BASED.value, FIModel.RFE_BASED.value ],
            'lightgbm': [FIModel.PERMUTATION_BASED.value,],
            'dummy': [FIModel.PERMUTATION_BASED.value],
        })

        model_to_methods = model_to_methods[['Reference', 'FI Methods']].set_index('Reference').squeeze().to_dict()
        
        model_to_methods.update({
            "sklearn.svm._classes.SVC": [FIModel.COEFFICIENT_BASED.value, FIModel.RFE_BASED.value],
        })
        
        self.model_to_methods = model_to_methods

        exclude = set(['dummy', 'svm', 'knn', 'nb', 'lightgbm', 'ada', 'gpc', 'rbfsvm'])
        all_models = set(self.clfx.models().index.tolist()) 
        external_models = [SVC(probability=True, kernel='linear'), ]

        include = list(all_models - exclude) + external_models
        
        if type(select_n_model) == float:
            self.learning_models = self.clfx.compare_models(n_select=round(len(self.clfx.models()) * select_n_model),
                                                             include=include,
                                                             )
        else:
            self.learning_models = self.clfx.compare_models(n_select=select_n_model,
                                                             include=include,
                                                             )
            
        return self

    def select_univariate_analysis_measurements(self, measurements=['Pearson', 'Spearman', 'Kendal']):
        self.measurements = measurements

        return self

    def fine_tune_selected_models(self, hpo_n_fold=4, hpo_n_iter=25, hpo_metric = 'Accuracy', hpo_search_library = 'scikit-optimize', hpo_search_algorithm = 'bayesian'):
        learning_models = self.learning_models
        model_to_methods = self.model_to_methods


        model_to_custom_config = {
            "sklearn.ensemble._forest.ExtraTreesClassifier": None,
            "sklearn.neighbors._classification.KNeighborsClassifier": None,
            "sklearn.ensemble._forest.RandomForestClassifier": None,
            "xgboost.sklearn.XGBClassifier": None,
            "lightgbm.sklearn.LGBMClassifier": {},
            "sklearn.neural_network._multilayer_perceptron.MLPClassifier": None,
            "sklearn.ensemble._gb.GradientBoostingClassifier": {},
            "sklearn.naive_bayes.GaussianNB": {},
            "sklearn.ensemble._weight_boosting.AdaBoostClassifier": {},
            "sklearn.linear_model._ridge.RidgeClassifier": {
                "alpha": [1],
                "solver": ["lsqr"],
            },
            "sklearn.discriminant_analysis.LinearDiscriminantAnalysis": {
                "shrinkage": [0.0,],
                "solver": ["lsqr"],
            },
            "sklearn.linear_model._logistic.LogisticRegression": {
                "C": [1,],
                "solver": ['liblinear'],
            },
            "sklearn.discriminant_analysis.QuadraticDiscriminantAnalysis": {},
            "sklearn.tree._classes.DecisionTreeClassifier": None,
            "sklearn.svm._classes.SVC": {
                'probability': [True, ],
                'C': [1.0],
                'kernel': [
                    "linear",
                ],
            },
        }

        tuned_learning_models = []
        for model in learning_models:
            print("-" * 80)
            print(model.__class__.__name__)

            if model_to_custom_config[full_name(model.__class__)] is None:
                tlm, tuner = self.clfx.tune_model(
                    model,
                    return_tuner=True,
                    fold=hpo_n_fold, 
                    verbose=True, 
                    tuner_verbose=True, 
                    optimize=hpo_metric, 
                    search_library=hpo_search_library, search_algorithm=hpo_search_algorithm,
                    n_iter=hpo_n_iter,
                    choose_better=True
                )

            elif len(model_to_custom_config[full_name(model.__class__)]) == 0:
                print("No hyper-tuning ...")
                tlm = model

            else:
                print("Custom config ...")
                tlm, tuner = self.clfx.tune_model(
                    model,
                    return_tuner=True,
                    fold=hpo_n_fold, 
                    verbose=True, 
                    tuner_verbose=True, 
                    optimize=hpo_metric, 
                    search_library='scikit-learn', search_algorithm='grid',
                    custom_grid=model_to_custom_config[full_name(model.__class__)], 
                    n_iter=hpo_n_iter,
                    choose_better=True
                )


            tuned_learning_models.append(tlm)

        self.finalized_tuned_learning_models = [self.clfx.finalize_model(estimator=tlm, model_only=True,) for tlm in tuned_learning_models]

        model_method_pairs = []

        for tlm in tuned_learning_models:
            methods = model_to_methods[full_name(tlm.__class__)]
            for method in methods:
                model_method_pairs.append((tlm, method))

        self.model_method_pairs = model_method_pairs
        
        return self

    def _initialize_model_feature_importance_data(self, models, n_iteration=5):
        feature_names = self.X.columns
        index_columns = ["Model Name", "Feature Importance Model", "Subset Index"]
        to_apply_product_list = [
            models,
            [i for i in range(n_iteration)],
        ]
        tuples = [(type(model).__name__, proper_fi_model, iteration) for (model, proper_fi_model), iteration in
                  itertools.product(*to_apply_product_list)]
        
        index = pd.MultiIndex.from_tuples(
            tuples,
            names=index_columns
        )

        data = pd.DataFrame({}, columns=feature_names, index=index)
        return data, index
    
    def _initialize_univariate_analysis_feature_importance_data(self, measurements, n_iteration=5):
        feature_names = self.X.columns
        index_columns = ["Model Name", "Feature Importance Model", "Subset Index"]
        to_apply_product_list = [
            measurements,
            [i for i in range(n_iteration)],
        ]
        tuples = [("Univariate Analysis", ua_measurement, iteration) for (ua_measurement), iteration in
                  itertools.product(*to_apply_product_list)]
        
        index = pd.MultiIndex.from_tuples(
            tuples,
            names=index_columns
        )

        data = pd.DataFrame({}, columns=feature_names, index=index)
        return data, index

    @staticmethod
    def stationary_bootstrap(X, y, n_iteration=5,):
        optimal_block_size = round(optimal_block_length(X).mean(axis=0)["stationary"])
        random_state = RandomState(1234)
        bootstraper = StationaryBootstrap(optimal_block_size, X=X, y=y, random_state=random_state)
        for bootstraped_data in bootstraper.bootstrap(n_iteration):
            yield bootstraped_data[1]['X'], bootstraped_data[1]['y']

    @staticmethod
    def iid_bootstrap(X, y, n_iteration=5,):
        random_state = RandomState(1234)
        bootstraper = IIDBootstrap(X=X, y=y, random_state=random_state)
        for bootstraped_data in bootstraper.bootstrap(n_iteration):
            yield bootstraped_data[1]['X'], bootstraped_data[1]['y']

    def compute_feature_importance_data(self, bootstrap_method, n_iteration=10, n_repeats=10):
        X = self.X
        y = self.y

        X = (X-X.mean()) / X.std()

        model_method_pairs = self.model_method_pairs
        models_fi, models_index = self._initialize_model_feature_importance_data(model_method_pairs, n_iteration=n_iteration)

        measurements = self.measurements
        measurements_fi, measurements_index = self._initialize_univariate_analysis_feature_importance_data(measurements, n_iteration=n_iteration)

        for iteration, (X_train, y_train) in tqdm(enumerate(
            bootstrap_method(X, y, n_iteration)
        ), total=n_iteration):

            for model, proper_fi_model in model_method_pairs:

                if proper_fi_model == FIModel.TREE_BASED.value:
                    try:
                        result = feature_importance_MDI(model, X_train, y_train, ).values
                    except:
                        result = feature_importance_sklearn(model, X_train, y_train,).values

                elif proper_fi_model == FIModel.PERMUTATION_BASED.value:
                    result = feature_importance_MDA(model, X_train, y_train, n_repeats=n_repeats).values

                elif proper_fi_model == FIModel.COEFFICIENT_BASED.value:
                    result = feature_importance_linear_models(model, X_train, y_train,).values

                elif proper_fi_model == FIModel.SINGLE_FEATURE_BASED.value:
                    result = feature_importance_SFI(model, X_train, y_train, n_splits=n_repeats).values
                
                elif proper_fi_model == FIModel.RFE_BASED.value:
                    result = feature_importance_RFE(model, X_train, y_train,).values

                # Normalize
                result = (result - result.min()) / (result.max() - result.min()) 

                models_fi.loc[type(model).__name__, proper_fi_model, iteration] = result

            ua_result = feature_importance_univariate_analysis(measurements, X_train, y_train).T.values
            measurements_fi.loc[pd.IndexSlice["Univariate Analysis", :, iteration]] = ua_result
            
        self.models_fi = models_fi
        self.models_index = models_index

        self.measurements_fi = measurements_fi
        self.measurements_index = measurements_index

        self.n_iterationn = n_iteration

        return self
    
    def compute_swefi_scores(self, percentage=0.5, weight='linear'):
        features = self.X.columns

        fi = pd.concat([self.models_fi, self.measurements_fi])
        index = self.models_index.tolist() + self.measurements_index.tolist()
        k = round(len(features) * percentage)
        important_features_together = []

        for idx in index:
            important_features_in_current_step = fi.loc[idx].squeeze().sort_values(ascending=False)[:k].index.tolist()
            important_features_together.extend(important_features_in_current_step)

        feature_ranked_as_k_top_important = dict(zip(features, [0] * len(features)))
        n_times_that_feature_ranked_as_important = Counter(important_features_together)
        dict.update(feature_ranked_as_k_top_important, n_times_that_feature_ranked_as_important)
        feature_ranked_as_k_top_important = pd.Series(feature_ranked_as_k_top_important)
        stability_scores = feature_ranked_as_k_top_important / (self.n_iterationn * (len(self.model_method_pairs) + len(self.measurements)))

        if weight == 'linear':
            weights = stability_scores
            weights = weights / weights.sum()
            swefi = (fi * weights)

        elif weight == 'logarithmic':
            weights = np.log(stability_scores + 1)
            weights = weights / weights.sum()
            swefi = (fi * weights)

        elif weight == 'exponential':
            weights = np.exp(stability_scores)
            weights = weights / weights.sum()
            swefi = (fi * weights) 

        elif weight == 'harmonic':
            swefi = (2 * fi * stability_scores) / (fi + stability_scores)

        elif weight == 'power-2':
            weights = stability_scores ** 2
            weights = weights / weights.sum()
            swefi = (fi * weights)

        elif weight == 'power-0.5':
            weights = (stability_scores ** 0.5)
            weights = weights / weights.sum()
            swefi = (fi * weights)

        elif weight == 'entropy':
            weights = (-stability_scores * np.log(stability_scores))
            weights = weights / weights.sum()            
            swefi = (fi * weights)

        else:
            raise NotImplementedError(f"Weight {weight} not implemented.")

        swefi = pd.concat(
            [swefi.mean() / swefi.mean().sum(), swefi.std() * (swefi.shape[0]) ** -0.5 / swefi.mean().sum()],
            axis=1,
        ).rename(columns={
            0: 'mean(SWEFI)',
            1: 'std(SWEFI)',
        }).sort_values(by='mean(SWEFI)')

        self.stability_scores = stability_scores
        self.swefi = swefi
        return self

    def get_swefi_scores(self):
        return self.swefi

    def get_inner_models_feature_importances(self):
        models_fi = self.models_fi.groupby(level=[0, 1]).sum()
        models_fi = models_fi.div(models_fi.sum(axis=1).values, axis=0).transpose()
        return models_fi


import pandas as pd
import numpy as np
from scipy.linalg import block_diag
from scipy.optimize import minimize
from sklearn.covariance import LedoitWolf
from sklearn.neighbors._kde import KernelDensity
from typing import Tuple, Union, Optional


def marcenko_pastur_pdf(
        variance: float,
        q: float,
        num_points: int
    ) -> pd.Series:
    """
    Computes the Marcenko-Pastur probability density function (pdf).

    :param variance: Variance of the observations
    :type variance: float
    :param q: Ratio T/N
    :type q: float
    :param num_points: Number of points in the pdf
    :type num_points: int
    :return: The Marcenko-Pastur pdf as a pandas Series
    :rtype: pd.Series

    The Marcenko-Pastur pdf is given by the formula:
    .. math::
       \frac{q}{{2 \pi \sigma \lambda}} \sqrt{(\lambda_{max} - \lambda)(\lambda - \lambda_{min})}

    where:
    - :math:`\lambda` is the eigenvalue
    - :math:`\sigma` is the variance of the observations
    - :math:`q` is the ratio T/N
    - :math:`\lambda_{max}` and :math:`\lambda_{min}` are the maximum and minimum eigenvalues respectively

    """
    lambda_min = variance * (1 - 1./q**0.5)**2
    lambda_max = variance * (1 + 1./q**0.5)**2
    eigenvalues = np.linspace(lambda_min, lambda_max, num_points).flatten()
    pdf = q / (2 * np.pi * variance * eigenvalues) * ((lambda_max - eigenvalues) * (eigenvalues - lambda_min))**0.5
    return pd.Series(pdf, index=eigenvalues)


def pca(
        matrix: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Computes the principal component analysis of a Hermitian matrix.

    :param matrix: Hermitian matrix
    :type matrix: np.ndarray
    :return: Eigenvalues and eigenvectors
    :rtype: Tuple[np.ndarray, np.ndarray]

    The principal component analysis is computed using the eigen decomposition of the Hermitian matrix.
    """
    eigenvalues, eigenvectors = np.linalg.eigh(matrix)
    indices = eigenvalues.argsort()[::-1]
    eigenvalues, eigenvectors = eigenvalues[indices], eigenvectors[:, indices]
    return np.diagflat(eigenvalues), eigenvectors

def fit_kde(
        observations: Union[np.ndarray, pd.Series],
        bandwidth: float = 0.25,
        kernel: str = 'gaussian',
        x: Optional[Union[np.ndarray, pd.Series]] = None
    ) -> pd.Series:
    """
    Fit a kernel density estimator to a series of observations.

    :param observations: Series of observations
    :type observations: Union[np.ndarray, pd.Series]
    :param bandwidth: Bandwidth of the kernel
    :type bandwidth: float
    :param kernel: Type of kernel to use (e.g., 'gaussian')
    :type kernel: str
    :param x: Array of values on which the fit KDE will be evaluated
    :type x: Optional[Union[np.ndarray, pd.Series]]
    :return: Kernel density estimate as a pandas Series
    :rtype: pd.Series
    """
    observations = np.asarray(observations).reshape(-1, 1)
    kde = KernelDensity(kernel=kernel, bandwidth=bandwidth).fit(observations)

    if x is None:
        x = np.unique(observations).reshape(-1, 1)
    else:
        x = np.asarray(x).reshape(-1, 1)

    log_prob = kde.score_samples(x)
    pdf = pd.Series(np.exp(log_prob), index=x.flatten())
    return pdf


def random_cov(
        num_columns: int,
        num_factors: int
    ) -> np.ndarray:
    """
    Generate a random covariance matrix.

    :param num_columns: Number of columns in the covariance matrix
    :type num_columns: int
    :param num_factors: Number of factors for random covariance matrix
    :type num_factors: int
    :return: Random covariance matrix
    :rtype: np.ndarray
    """
    w = np.random.normal(size=(num_columns, num_factors))
    cov = np.dot(w, w.T)
    cov += np.diag(np.random.uniform(size=num_columns))
    return cov


def cov_to_corr(cov: np.ndarray) -> np.ndarray:
    """
    Convert a covariance matrix to a correlation matrix.

    :param cov: Covariance matrix
    :type cov: np.ndarray
    :return: Correlation matrix
    :rtype: np.ndarray
    """
    std = np.sqrt(np.diag(cov))
    corr = cov / np.outer(std, std)
    corr = np.clip(corr, -1, 1)  # fix numerical errors
    return corr

def error_pdfs(
        variance: float,
        eigenvalues: np.ndarray,
        q: float,
        bandwidth: float,
        num_points: int = 1000
    ) -> float:
    """
    Computes the sum of squared errors between the theoretical and empirical PDFs.

    :param variance: Variance of the observations
    :type variance: float
    :param eigenvalues: Eigenvalues of the correlation matrix
    :type eigenvalues: np.ndarray
    :param q: Ratio T/N
    :type q: float
    :param bandwidth: Bandwidth of the kernel
    :type bandwidth: float
    :param num_points: Number of points in the PDF
    :type num_points: int
    :return: Sum of squared errors between the theoretical and empirical PDFs
    :rtype: float
    """
    pdf0 = marcenko_pastur_pdf(variance, q, num_points)
    pdf1 = fit_kde(eigenvalues, bandwidth, x=pdf0.index.values)
    sse = np.sum((pdf1 - pdf0)**2)
    return sse


def find_max_eval(
        eigenvalues: np.ndarray,
        q: float,
        bandwidth: float
    ) -> Tuple[float, float]:
    """
    Find the maximum random eigenvalue by fitting the Marcenko-Pastur distribution.

    :param eigenvalues: Eigenvalues of the correlation matrix
    :type eigenvalues: np.ndarray
    :param q: Ratio T/N
    :type q: float
    :param bandwidth: Bandwidth of the kernel
    :type bandwidth: float
    :return: Maximum random eigenvalue and its variance
    :rtype: Tuple[float, float]
    """
    result = minimize(lambda *x: error_pdfs(*x), 0.5, args=(eigenvalues, q, bandwidth), bounds=((1E-5, 1-1E-5),))
    variance = result['x'][0] if result['success'] else 1
    emax = variance * (1 + (1./q)**0.5)**2
    return emax, variance


def denoised_corr(
        eigenvalues: np.ndarray,
        eigenvectors: np.ndarray,
        num_factors: int
    ) -> np.ndarray:
    """
    Remove noise from the correlation matrix by fixing random eigenvalues.

    :param eigenvalues: Eigenvalues of the correlation matrix
    :type eigenvalues: np.ndarray
    :param eigenvectors: Eigenvectors of the correlation matrix
    :type eigenvectors: np.ndarray
    :param num_factors: Number of factors for the correlation matrix
    :type num_factors: int
    :return: Denoised correlation matrix
    :rtype: np.ndarray
    """
    eval_copy = np.diag(eigenvalues).copy()
    eval_copy[num_factors:] = eval_copy[num_factors:].sum() / float(eval_copy.shape[0] - num_factors)
    eval_copy = np.diag(eval_copy)
    corr1 = np.dot(eigenvectors, eval_copy).dot(eigenvectors.T)
    corr1 = cov_to_corr(corr1)
    return corr1


def denoised_corr2(
        eigenvalues: np.ndarray,
        eigenvectors: np.ndarray,
        num_factors: int,
        alpha: float = 0
    ) -> np.ndarray:
    """
    Remove noise from the correlation matrix through targeted shrinkage.

    :param eigenvalues: Eigenvalues of the correlation matrix
    :type eigenvalues: np.ndarray
    :param eigenvectors: Eigenvectors of the correlation matrix
    :type eigenvectors: np.ndarray
    :param num_factors: Number of factors for the correlation matrix
    :type num_factors: int
    :param alpha: Shrinkage parameter
    :type alpha: float
    :return: Denoised correlation matrix
    :rtype: np.ndarray
    """
    eval_L, evec_L = eigenvalues[:num_factors, :num_factors], eigenvectors[:, :num_factors]
    eval_R, evec_R = eigenvalues[num_factors:, num_factors:], eigenvectors[:, num_factors:]
    corr0 = np.dot(evec_L, eval_L).dot(evec_L.T)
    corr1 = np.dot(evec_R, eval_R).dot(evec_R.T)
    corr2 = corr0 + alpha*corr1 + (1 - alpha)*np.diag(np.diag(corr1))
    return corr2

def form_block_matrix(
        n_blocks: int,
        block_size: int,
        block_correlation: float
    ) -> np.ndarray:
    """
    Forms a block diagonal correlation matrix.

    :param n_blocks: Number of blocks
    :type n_blocks: int
    :param block_size: Size of each block
    :type block_size: int
    :param block_correlation: Correlation within each block
    :type block_correlation: float
    :return: Block diagonal correlation matrix
    :rtype: np.ndarray
    """
    block = np.ones((block_size, block_size)) * block_correlation
    block[range(block_size), range(block_size)] = 1
    corr = block_diag(*([block] * n_blocks))
    return corr

def form_true_matrix(
        n_blocks: int,
        block_size: int,
        block_correlation: float
    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Forms a shuffled block diagonal correlation matrix and the corresponding covariance matrix.

    :param n_blocks: Number of blocks
    :type n_blocks: int
    :param block_size: Size of each block
    :type block_size: int
    :param block_correlation: Correlation within each block
    :type block_correlation: float
    :return: Mean and covariance matrix
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    corr0 = form_block_matrix(n_blocks, block_size, block_correlation)
    corr0 = pd.DataFrame(corr0)
    cols = corr0.columns.tolist()
    np.random.shuffle(cols)
    corr0 = corr0[cols].loc[cols].copy(deep=True)
    std0 = np.random.uniform(.05, .2, corr0.shape[0])
    cov0 = corr_to_cov(corr0, std0)
    mu0 = np.random.normal(std0, std0, cov0.shape[0]).reshape(-1, 1)
    return mu0, cov0

def simulates_cov_mu(
        mu0: np.ndarray,
        cov0: np.ndarray,
        n_obs: int,
        shrink: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Simulates multivariate normal observations and computes the sample mean and covariance.

    :param mu0: True mean
    :type mu0: np.ndarray
    :param cov0: True covariance matrix
    :type cov0: np.ndarray
    :param n_obs: Number of observations
    :type n_obs: int
    :param shrink: Whether to use Ledoit-Wolf shrinkage
    :type shrink: bool
    :return: Sample mean and covariance matrix
    :rtype: Tuple[np.ndarray, np.ndarray]
    """
    x = np.random.multivariate_normal(mu0.flatten(), cov0, size=n_obs)
    mu1 = x.mean(axis=0).reshape(-1, 1)
    cov1 = LedoitWolf().fit(x).covariance_ if shrink else np.cov(x, rowvar=0)
    return mu1, cov1

def corr_to_cov(
        corr: np.ndarray,
        std: np.ndarray
    ) -> np.ndarray:
    """
    Converts a correlation matrix to a covariance matrix.

    :param corr: Correlation matrix
    :type corr: np.ndarray
    :param std: Standard deviations
    :type std: np.ndarray
    :return: Covariance matrix
    :rtype: np.ndarray
    """
    return corr * np.outer(std, std)

def denoise_cov(
        cov0: np.ndarray,
        q: float,
        bandwidth: float
    ) -> np.ndarray:
    """
    De-noises the covariance matrix.

    :param cov0: Covariance matrix
    :type cov0: np.ndarray
    :param q: Ratio of number of observations to number of variables
    :type q: float
    :param bandwidth: Bandwidth parameter
    :type bandwidth: float
    :return: De-noised covariance matrix
    :rtype: np.ndarray
    """
    corr0 = cov_to_corr(cov0)
    eval0, evec0 = np.linalg.eigh(corr0)
    eval0 = np.diag(eval0)
    emax0, var0 = find_max_eval(eval0, q, bandwidth)
    n_facts0 = eval0.shape[0] - np.diag(eval0)[::-1].searchsorted(emax0)
    corr1 = denoised_corr(eval0, evec0, n_facts0)
    cov1 = corr_to_cov(corr1, np.diag(cov0)**0.5)
    return cov1

def optimal_portfolio(
        cov: np.ndarray,
        mu: np.ndarray = None
    ) -> np.ndarray:
    """
    Computes the optimal portfolio weights.

    :param cov: Covariance matrix
    :type cov: np.ndarray
    :param mu: Expected returns
    :type mu: np.ndarray
    :return: Optimal portfolio weights
    :rtype: np.ndarray
    """
    inv = np.linalg.inv(cov)
    ones = np.ones(shape=(inv.shape[0], 1))
    if mu is None:
        mu = ones
    w = np.dot(inv, mu)
    w /= np.dot(ones.T, w)
    return w



import imp
import numpy as np
import pandas as pd 
from numba import jit
from RiskLabAI.hpc import *
from scipy.stats import norm

def probability_bet_size(
    probabilities: np.ndarray,
    sides: np.ndarray
) -> np.ndarray:
    """
    Calculate the bet size based on probabilities and side.

    :param probabilities: array of probabilities
    :param sides: array indicating the side of the bet (e.g., long/short or buy/sell)
    :return: array of bet sizes

    .. math::

       \text{bet size} = \text{side} \times (2 \times \text{CDF}(\text{probabilities}) - 1)
    """
    return sides * (2 * norm.cdf(probabilities) - 1)

@jit(nopython=True)
def average_bet_sizes(
    price_dates: np.ndarray,
    start_dates: np.ndarray,
    end_dates: np.ndarray,
    bet_sizes: np.ndarray
) -> np.ndarray:
    """
    Compute average bet sizes for each date.

    :param price_dates: array of price dates
    :param start_dates: array of start dates for bets
    :param end_dates: array of end dates for bets
    :param bet_sizes: array of bet sizes for each date range
    :return: array of average bet sizes for each price date
    """
    num_dates = len(price_dates)
    avg_bet_sizes = np.zeros(num_dates)
    
    for i in range(num_dates):
        total = 0
        count = 0
        for j in range(len(start_dates)):
            if start_dates[j] <= price_dates[i] <= end_dates[j]:
                total += bet_sizes[j]
                count += 1
        if count > 0:
            avg_bet_sizes[i] = total / count
            
    return avg_bet_sizes

def strategy_bet_sizing(
    price_timestamps: pd.Series,
    times: pd.Series,
    sides: pd.Series,
    probabilities: pd.Series
) -> pd.Series:
    """
    Calculate the average bet size for a trading strategy given price timestamps.

    :param price_timestamps: series of price timestamps
    :param times: series with start times as indices and end times as values
    :param sides: series indicating the side of the position (e.g., long/short)
    :param probabilities: series of probabilities associated with each position
    :return: series of average bet sizes for each price timestamp
    """
    
    bet_sizes = probability_bet_size(probabilities.to_numpy(), sides.to_numpy())
    
    avg_bet_sizes = average_bet_sizes(price_timestamps.to_numpy(), times.index.to_numpy(), times.values, bet_sizes)

    return pd.Series(avg_bet_sizes, index=price_timestamps)

"""----------------------------------------------------------------------
    function: Calculation of Average Active Signals 
    reference: De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
    methodology: SNIPPET 10.2
----------------------------------------------------------------------"""
def avgActiveSignals(signals,# DataFrame of signal
                     nThreads): #number of threads 
    # compute the average signal among those active
    #1) time points where signals change (either one starts or one ends)
    timePoints=set(signals['t1'].dropna().values) #drop nan value from dataframe
    timePoints=timePoints.union(signals.index.values) #union incex 
    timePoints=list(timePoints) #list them!
    timePoints.sort() #sort them!
    out=mpPandasObj(mpAvgActiveSignals,('molecule',timePoints),nThreads,signals=signals) #generate final signal
    return out
#———————————————————————————————————————

"""
function: Calculation of Average Active Signals 
reference: De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
methodology: SNIPPET 10.2
At time loc, average signal among those still active.
Signal is active if:
a) issued before or at loc AND
b) loc before signal's endtime, or endtime is still unknown (NaT).
"""
def mpAvgActiveSignals(signals, #DataFrame of signal 
                       molecule): #index of dataFrame that function act on it 
    out=pd.Series()
    for loc in molecule:
        signal_=(signals.index.values<=loc)&((loc<signals['t1'])|pd.isnull(signals['t1'])) #keep signal that contain loc
        act=signals[signal_].index #store index 
        if len(act)>0:
            out[loc]=signals.loc[act,'signal'].mean() #get mean!
        else:
            out[loc]=0 # no signals active at this time
    return out

"""----------------------------------------------------------------------
    function: Discretize Signals 
    reference: De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
    methodology: SNIPPET 10.3
----------------------------------------------------------------------"""
def discreteSignal(signal, #DataFrame of signal  
                   stepSize): #stepsize for dicretize signal 
    # discretize signal
    discreteSignal=(signal/stepSize).round()*stepSize # discretize
    discreteSignal[discreteSignal>1]=1 # cap
    discreteSignal[discreteSignal<-1]=-1 # floor
    return discreteSignal

"""----------------------------------------------------------------------
    function: generate Signal 
    reference: De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
    methodology: SNIPPET 10.1
----------------------------------------------------------------------"""
def Signal(events, #DataFrame of events that genrate in labelling
           stepSize, #Step size for discretize signal 
           probability, #probability that come from meta labelling 
           prediction, # prediction of primary model 
           nClasses, #number of class of prediction  
           nThreads,): # number of threads  # o
    # get signals from predictions
    if probability.shape[0]==0:return pd.Series()
    #1) generate signals from multinomial classification (one-vs-rest, OvR)
    discreteSignal=(probability-1./nClasses)/(probability*(1.-probability))**.5 # t-value of OvR
    newSignal=prediction*(2*norm.cdf(discreteSignal)-1) # signal=side*size
    if 'side' in events:newSignal*=events.loc[newSignal.index,'side'] # meta-labeling
    #2) compute average signal among those concurrently open
    newSignal=newSignal.to_frame('signal').join(events[['t1']],how='left')
    newSignal=avgActiveSignals(newSignal,nThreads) #average signal that contain one events 
    newSignal=discreteSignal(signal=newSignal,stepSize=stepSize) #discretize  signal 
    return newSignal

"""----------------------------------------------------------------------
    function: DYNAMIC POSITION SIZE AND LIMIT PRICE 
    reference: De Prado, M. (2018) Advances in financial machine learning. John Wiley & Sons.
    methodology: p.145 SNIPPET 10.4 
----------------------------------------------------------------------"""
def betSize(w,x):
    return x*(w + x**2)**-.5
#———————————————————————————————————————
def TPos(w, # coefficient that regulates the width of the sigmoid function
         f, # predicted price
         acctualPrice , # acctual price 
         maximumPositionSize): #  maximum absolute position size
    return int(betSize(w, f-acctualPrice)*maximumPositionSize)
#———————————————————————————————————————
def inversePrice(f, # predicted price
                 w, #coefficient that regulates the width of the sigmoid function
                 m): # betsize 
    return f - m*(w/(1-m^2))^(.5)
#———————————————————————————————————————
def limitPrice(targetPositionSize, #  target position size
               cPosition, # current position
               f, #  predicted price
               w, # coefficient that regulates the width of the sigmoid function
               maximumPositionSize): # maximum absolute position size
    if targetPositionSize >=  cPosition:
        sgn = 1
    else:
        sgn = -1
    lP = 0
    for i in range(abs(cPosition+sgn),abs(targetPositionSize)) :
        lP += inversePrice(f,w,i/float(maximumPositionSize))
    lP /= targetPositionSize-cPosition
    return lP    
#———————————————————————————————————————
def getW(x, # divergence between the current market price and the forecast
         m): # bet size 
    return x**2*(1/m**(-2)-1)


# from feature_importance_strategy import FeatureImportanceStrategy
from RiskLabAI.features.feature_importance.feature_importance_strategy import FeatureImportanceStrategy
import numpy as np
import pandas as pd
from sklearn.metrics import log_loss
from sklearn.model_selection import KFold
from typing import List, Optional


class FeatureImportanceMDA(FeatureImportanceStrategy):
    """
    Computes the feature importance using the Mean Decrease Accuracy (MDA) method.

    The method shuffles each feature one by one and measures how much the performance 
    (log loss in this context) decreases due to the shuffling.

    .. math::

        \\text{importance}_{j} = \\frac{\\text{score without shuffling} - \\text{score with shuffling}_{j}}
        {\\text{score without shuffling}}

    """

    def __init__(
            self,
            classifier: object,
            x: pd.DataFrame,
            y: pd.Series,
            n_splits: int = 10,
            score_sample_weights: Optional[List[float]] = None,
            train_sample_weights: Optional[List[float]] = None
    ) -> None:
        """
        Initialize the class with parameters.

        :param classifier: The classifier object.
        :param x: The feature data.
        :param y: The target data.
        :param n_splits: Number of splits for cross-validation.
        :param score_sample_weights: Weights for scoring samples.
        :param train_sample_weights: Weights for training samples.
        """
        self.classifier = classifier
        self.x = x
        self.y = y
        self.n_splits = n_splits
        self.score_sample_weights = score_sample_weights
        self.train_sample_weights = train_sample_weights

    def compute(self) -> pd.DataFrame:
        """
        Compute the feature importances.

        :return: Feature importances as a dataframe with "Mean" and "StandardDeviation" columns.
        """
        if self.train_sample_weights is None:
            self.train_sample_weights = np.ones(self.x.shape[0])
        if self.score_sample_weights is None:
            self.score_sample_weights = np.ones(self.x.shape[0])

        cv_generator = KFold(n_splits=self.n_splits)
        initial_scores, shuffled_scores = pd.Series(dtype=float), pd.DataFrame(columns=self.x.columns)

        for i, (train, test) in enumerate(cv_generator.split(self.x)):
            print(f"Fold {i} start ...")

            x_train, y_train, weights_train = self.x.iloc[train, :], self.y.iloc[train], self.train_sample_weights[train]
            x_test, y_test, weights_test = self.x.iloc[test, :], self.y.iloc[test], self.score_sample_weights[test]

            fitted_classifier = self.classifier.fit(X=x_train, y=y_train, sample_weight=weights_train)
            prediction_probability = fitted_classifier.predict_proba(x_test)

            initial_scores.loc[i] = -log_loss(
                y_test,
                prediction_probability,
                labels=self.classifier.classes_,
                sample_weight=weights_test
            )

            for feature in self.x.columns:
                x_test_shuffled = x_test.copy(deep=True)
                np.random.shuffle(x_test_shuffled[feature].values)
                shuffled_proba = fitted_classifier.predict_proba(x_test_shuffled)
                shuffled_scores.loc[i, feature] = -log_loss(y_test, shuffled_proba, labels=self.classifier.classes_)

        importances = (-1 * shuffled_scores).add(initial_scores, axis=0)
        importances /= (-1 * shuffled_scores)

        importances = pd.concat({
            "Mean": importances.mean(),
            "StandardDeviation": importances.std() * importances.shape[0]**-0.5
        }, axis=1)

        return importances



from . import validation
from .backtest_statistics import *
from .backtest_synthetic_data import *
from .bet_sizing import *
from .strategy_risk import *
from .test_set_overfitting import *
from .probability_of_backtest_overfitting import *
from .probabilistic_sharpe_ratio import *
from .backtset_overfitting_in_the_machine_learning_era_simulation import *



import numpy as np
from numba import jit
from numba import float64
from numba import int64

@jit((float64[:], int64), nopython=False, nogil=True)
def ewma(
    array, 
    window
):
    """
    This function calculate Exponential Weighted Moving Average of array
    :param array: input array
    :param window: window size
    :return: ewma array
    """
    length = array.shape[0]
    result_ewma_array = np.empty(length, dtype=float64)

    α = 2 / (window + 1)
    weight = 1
    current_ewma_term = array[0]
    result_ewma_array[0] = current_ewma_term
    for i in range(1, length):
        weight += (1 - α) ** i
        current_ewma_term = current_ewma_term * (1 - α) + array[i]
        result_ewma_array[i] = current_ewma_term / weight

    return result_ewma_array


from typing import Dict, Union, Tuple, List, Any, Optional, Generator
import pandas as pd
import numpy as np
from abc import ABC, abstractmethod

class CrossValidator(ABC):
    """
    Abstract Base Class (ABC) for cross-validation strategies.
    Handles both single data inputs and dictionary inputs.

    :param data: The input data, either as a single DataFrame or a dictionary of DataFrames.
    :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]

    :param labels: The labels corresponding to the data, either as a single Series or a dictionary of Series.
    :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]

    :param groups: Optional group labels for stratified splitting.
    :type groups: Optional[np.ndarray]
    """


    @abstractmethod
    def get_n_splits(
        self,
        data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]] = None,
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> int:
        """
        Return number of splits.

        :param data: Dataset or dictionary of datasets.
        :type data: Optional[Union[pd.DataFrame, Dict[str, pd.DataFrame]]]

        :param labels: Labels or dictionary of labels.
        :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]

        :param groups: Group labels for the samples.
        :type groups: Optional[np.ndarray]

        :return: Number of splits.
        :rtype: int
        """


    @abstractmethod
    def _single_split(
        self,
        single_data: pd.DataFrame,
    ) -> Generator[Tuple[np.ndarray, np.ndarray], None, None]:
        """
        Splits a single data set into train-test indices.

        This function provides train-test indices to split the data into train/test sets
        by respecting the time order (if applicable) and the specified number of splits.

        :param single_data: Input dataset.
        :type single_data: pd.DataFrame

        :return: Generator yielding train-test indices.
        :rtype: Generator[Tuple[np.ndarray, np.ndarray], None, None]
        """


    @abstractmethod
    def split(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Optional[Union[pd.Series, Dict[str, pd.Series]]] = None,
        groups: Optional[np.ndarray] = None
    ) -> Union[
        Generator[Tuple[np.ndarray, np.ndarray], None, None],
        Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]
    ]:
        """
        Splits data or a dictionary of data into train-test indices.

        This function returns a generator that yields train-test indices. If a dictionary
        of data is provided, the generator yields a key followed by the train-test indices.

        :param data: Dataset or dictionary of datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Labels or dictionary of labels.
        :type labels: Optional[Union[pd.Series, Dict[str, pd.Series]]]
        :param groups: Group labels for the samples.
        :type groups: Optional[np.ndarray]

        :return: Generator yielding either train-test indices directly or a key
                followed by train-test indices.
        :rtype: Union[
            Generator[Tuple[np.ndarray, np.ndarray], None, None],
            Generator[Tuple[str, Tuple[np.ndarray, np.ndarray]], None, None]
        ]
        """


    @abstractmethod
    def _single_backtest_paths(
        self,
        single_data: pd.DataFrame
    ) -> Dict[str, List[Dict[str, List[np.ndarray]]]]:
        """
        Generates backtest paths for a single dataset.

        This function creates and returns backtest paths (i.e., combinations of training and test sets)
        for a single dataset by applying k-fold splitting or any other splitting strategy defined
        by the `_single_split` function.

        :param single_data: Input dataset.
        :type single_data: pd.DataFrame

        :return: Dictionary of backtest paths.
        :rtype: Dict[str, List[Dict[str, List[np.ndarray]]]]
        """


    @abstractmethod
    def backtest_paths(
        self,
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
    ) -> Union[
        Dict[str, List[Dict[str, np.ndarray]]],
        Dict[str, Dict[str, List[Dict[str, List[np.ndarray]]]]]
    ]:
        """
        Generates backtest paths for data.

        This function returns backtest paths for either a single dataset or a dictionary
        of datasets. Each backtest path consists of combinations of training and test sets.

        :param data: Dataset or dictionary of datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Labels or dictionary of labels.
        :type labels: Union[pd.Series, Dict[str, pd.Series]]

        :return: Dictionary of backtest paths or dictionary of dictionaries for multiple datasets.
        :rtype: Union[
            Dict[str, List[Dict[str, np.ndarray]]],
            Dict[str, Dict[str, List[Dict[str, List[np.ndarray]]]]]
        ]
        """


    @abstractmethod
    def _single_backtest_predictions(
        self,
        single_estimator: Any,
        single_data: pd.DataFrame,
        single_labels: pd.Series,
        single_weights: Optional[np.ndarray] = None,
        n_jobs: int = 1
    ) -> Dict[str, np.ndarray]:
        """
        Obtain predictions for a single dataset during backtesting.

        This function leverages parallel computation to train and predict on different train-test splits
        of a single dataset using a given estimator. It utilizes the `_single_split` method to generate
        the train-test splits.

        :param single_estimator: Estimator or model to be trained and used for predictions.
        :type single_estimator: Any
        :param single_data: Data of the single dataset.
        :type single_data: pd.DataFrame
        :param single_labels: Labels corresponding to the single dataset.
        :type single_labels: pd.Series
        :param single_weights: Weights for the observations in the single dataset.
                            Defaults to equally weighted if not provided.
        :type single_weights: np.ndarray, optional
        :param n_jobs: The number of jobs to run in parallel. Default is 1.
        :type n_jobs: int, optional
        :return: Predictions structured in a dictionary for the backtest paths.
        :rtype: Dict[str, np.ndarray]
        """

    @abstractmethod
    def backtest_predictions(
        self,
        estimator: Union[Any, Dict[str, Any]],
        data: Union[pd.DataFrame, Dict[str, pd.DataFrame]],
        labels: Union[pd.Series, Dict[str, pd.Series]],
        sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]] = None,
        predict_probability: bool = False,
        n_jobs: int = 1
    ) -> Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]:
        """
        Generate backtest predictions for single or multiple datasets.

        For each dataset, this function leverages the `_single_backtest_predictions` method to obtain
        predictions for different train-test splits using the given estimator.

        :param estimator: Model or estimator to be trained and used for predictions.
                        Can be a single estimator or a dictionary of estimators for multiple datasets.
        :type estimator: Union[Any, Dict[str, Any]]
        :param data: Input data for training and testing. Can be a single dataset or
                    a dictionary of datasets for multiple datasets.
        :type data: Union[pd.DataFrame, Dict[str, pd.DataFrame]]
        :param labels: Target labels for training and testing. Can be a single series or
                    a dictionary of series for multiple datasets.
        :type labels: Union[pd.Series, Dict[str, pd.Series]]
        :param sample_weights: Weights for the observations in the dataset(s).
                            Can be a single array or a dictionary of arrays for multiple datasets.
                            Defaults to None, which means equal weights for all observations.
        :type sample_weights: Optional[Union[np.ndarray, Dict[str, np.ndarray]]]
        :param n_jobs: The number of jobs to run in parallel. Default is 1.
        :type n_jobs: int, optional
        :return: Backtest predictions structured in a dictionary (or nested dictionaries for multiple datasets).
        :rtype: Union[Dict[str, np.ndarray], Dict[str, Dict[str, np.ndarray]]]
        """



import torch
import torch.nn as nn
import torch.autograd as autograd
import numpy as np

from RiskLabAI.pde.model import *
from RiskLabAI.pde.equation import *
def initialize_weights(m: nn.Module) -> None:
    """
    Initializes the weights of the given module.

    Args:
    - m (nn.Module): the module to initialize weights of

    Returns:
    - None
    """
    if isinstance(m, nn.Linear):
        torch.nn.init.normal_(m.weight, 0, 0.1)
        m.bias.data.fill_(0.00)


class FBSDESolver:
    def __init__(
            self,
            pde,
            layer_sizes,
            learning_rate,
            solving_method, 
            device
    ):
        """
        Initializes the FBSDESolver.

        Args:
        - pde : the partial differential equation to solve
        - layer_sizes (list[int]): list of sizes of hidden layers
        - learning_rate (float): learning rate for optimization
        - solving_method (str): method to solve the PDE ('Monte-Carlo', 'Deep-Time-SetTransformer', 'Basic')
        """
        self.pde = pde
        self.layer_size = layer_sizes
        self.learning_rate = learning_rate
        self.method = solving_method
        self.device = device
        if solving_method == 'Monte-Carlo':
            self.solver = TimeDependentNetworkMonteCarlo(pde.dim, self.layer_size, pde.dim, pde.sigma)
        elif solving_method == 'Deep-Time-SetTransformer':
            self.solver = DeepTimeSetTransformer(1)
        elif solving_method == 'DTNN':
            self.solver = TimeDependentNetwork(pde.dim, self.layer_size, pde.dim).to(self.device)
        elif solving_method == 'DeepBSDE':
            self.solver = []
            for i in range(self.pde.num_time_interval):
              self.solver.append(DeepBSDE(layer_sizes).to(self.device))



        if solving_method != 'DeepBSDE':
          self.optimizer = torch.optim.Adam(self.solver.parameters(), lr=self.learning_rate,betas=(0.9, 0.99))
        else  :
          self.optimizer = torch.optim.Adam((par for model in self.solver for par in model.parameters()), lr=self.learning_rate, betas=(0.9, 0.99))

    def compute_loss(self,
                     y,
                     dw,
                     t,
                     init,
                     init_grad
                     ):

      batch_size = y.size()[0]
      y_terminal = torch.ones((batch_size,)).to(self.device) * init.to(self.device)

      coef = torch.ones((batch_size,)).to(self.device)
      dw_coef = torch.zeros((batch_size,)).to(self.device)
      L1 = 0.0
      if self.method == 'Monte-Carlo':
          num_sample = 5000
          primary_sample = torch.unsqueeze(torch.randn(size=[num_sample, self.pde.dim]), dim=0).to(self.device)
      for z in range(self.pde.num_time_interval):
          S0 = y[:, :, z]

          t0 = t * self.pde.delta_t * z
          if self.method == 'Monte-Carlo':
              ders = []
              time_length = (self.pde.num_time_interval - z) * self.pde.delta_t
              current_state = torch.unsqueeze(S0, dim=1).to(self.device)
              dw_sample = primary_sample * torch.tensor(np.sqrt(time_length)).to(self.device)
              k = (1 + self.pde.mu_bar * time_length) * current_state + (self.pde.sigma * current_state * dw_sample)
              ddw = dw_sample
              means = torch.mean((self.pde.terminal_for_sample(k)), dim=1, keepdim=True)
              mins = torch.mean((self.pde.terminal_for_sample(k) - means) * ddw, dim=1) / torch.tensor(
                  (self.pde.num_time_interval - z) * self.pde.delta_t).to(self.device)
              ders = torch.tensor(mins).to(torch.float32).to(self.device)
              out1 = self.solver(t0, S0, ders)
          elif self.method == 'Deep-Time-SetTransformer':
              t = torch.ones((batch_size, 1, 1)).to(self.device)
              t0 = t * self.pde.delta_t * z
              S0 = S0.reshape(S0.size()[0], 100, 1).requires_grad_(True)
              out1 = self.solver(t0, S0)
              out1 = torch.squeeze(autograd.grad(out1.sum(), S0, create_graph=True)[0])

              S0 = torch.squeeze(S0)

              out1 = out1 * self.sigma_matrix(S0)
          elif self.method == 'DTNN':
              out1 = self.solver(t0, S0)
          elif self.method == 'DeepBSDE':
              if z > 0 :
                out1 = self.solver[z](S0)
                #out1 = torch.zeros((S0.size()[0] , self.pde.dim)).to(device)
              else :
                out1 = init_grad * torch.ones((S0.size()[0] , self.pde.dim)).to(self.device)
                #out1 = torch.zeros((S0.size()[0] , self.pde.dim)).to(device)



          else:
              print('Model type is not true')

          samp = dw[:, :, z].to(self.device)

          nsim = S0.size()[0]
          interest_rate = torch.squeeze(self.pde.r_u(t0, S0, y_terminal, out1).to(self.device))
          hz = self.pde.h_z(t0, S0, y_terminal, out1)
          if z > 0:
              dw_coef = dw_coef * (1 + interest_rate * self.pde.delta_t) + torch.squeeze(
                  torch.bmm(torch.unsqueeze(out1, 1), torch.unsqueeze(samp, 2))) + torch.squeeze(
                  self.pde.h_z(t0, S0, y_terminal, out1)).to(self.device) * self.pde.delta_t
          else:
              dw_coef = torch.squeeze(
                  torch.bmm(torch.unsqueeze(out1, 1), torch.unsqueeze(samp, 2))) + torch.squeeze(
                  self.pde.h_z(t0, S0, y_terminal, out1)).to(self.device) * self.pde.delta_t
          coef = coef * (1 + interest_rate * self.pde.delta_t)
          y_terminal = y_terminal * (1 + interest_rate * self.pde.delta_t) + torch.squeeze(
              self.pde.h_z(t0, S0, y_terminal, out1).to(self.device)) * self.pde.delta_t + torch.squeeze(
              torch.bmm(torch.unsqueeze(out1, 1), torch.unsqueeze(samp, 2)))


      S0 = y[:, :, self.pde.num_time_interval]
      t0 = t * self.pde.delta_t * self.pde.num_time_interval

      payoff = self.pde.terminal(t0, S0)

      coef = coef.to(self.device)



      loss = torch.mean(torch.square(torch.squeeze(payoff) - torch.squeeze(y_terminal)))
      return loss,coef,dw_coef,payoff
    def solve(
            self,
            num_iterations,
            batch_size,
            init,
            sample_size=None
    ):
        """
        Solves the PDE.

        Args:
        - num_iterations (int): number of iterations for optimization
        - batch_size (int): batch size for training
        - init (torch.Tensor): initial value
        - device (torch.device): device to perform calculations on ('cpu', 'cuda')
        - sample_size (int, optional): sample size for Monte-Carlo method

        Returns:
        - list[torch.Tensor]: list of losses during optimization
        - list[torch.Tensor]: list of initial values during optimization
        """
        losses = []
        inits = []


        torch.seed()
        init_grad = (torch.zeros(1,self.pde.dim).to(self.device)).to(self.device).requires_grad_()
        if self.method != 'DeepBSDE':
          self.solver.train()
        else :
          for i in range(self.pde.num_time_interval):
            self.solver[i].train()
          init =  (torch.ones(1).to(self.device) *init).to(self.device).requires_grad_()
          init_grad = (torch.zeros(1,self.pde.dim).to(self.device)).to(self.device).requires_grad_()
          init_opt = torch.optim.Adam([init],lr=0.1, betas=(0.9, 0.99))
          init_grad_opt = torch.optim.Adam([init_grad],lr=0.001, betas=(0.9, 0.99))


        dw_val, y_val = self.pde.sample(128)
        y_val = torch.tensor(y_val)
        y_val = y_val.to(torch.float32).to(self.device)
        dw_val = torch.tensor(dw_val).to(torch.float32)
        dw_val = dw_val.to(self.device)
        t_val = torch.ones((128, 1)).to(self.device)


        for j in range(num_iterations):
            print(j + 1)

            dw, y = self.pde.sample(batch_size)
            y = torch.tensor(y)
            y = y.to(torch.float32)
            dw = torch.tensor(dw).to(torch.float32)
            dw = dw.to(self.device)
            t = torch.ones((batch_size, 1)).to(self.device)

            self.optimizer.zero_grad()
            if self.method  == 'DeepBSDE' :
              init_opt.zero_grad()
              init_grad_opt.zero_grad()


            y = y.to(self.device)

            loss,coef,dw_coef,payoff = self.compute_loss(y,dw,t,init , init_grad)





            loss.backward()

            self.optimizer.step()



            loss,coef,dw_coef,payoff = self.compute_loss(y_val,dw_val,t_val,init , init_grad)
            #print(loss)

            inits.append(init.detach().cpu().numpy())
            print(init.detach().cpu().numpy())
            if self.method != 'DeepBSDE' :
              init = (torch.mean((torch.squeeze(payoff) - dw_coef.detach()) / coef).detach())
            else :
              init_opt.step()
              init_grad_opt.step()

            #print(init)

            losses.append(loss.detach().cpu().numpy())
            print(loss.detach().cpu().numpy())


        return losses, inits



class FBSNNolver:
    def __init__(
            self,
            pde,
            layer_sizes,
            learning_rate,
            device
    ):
        """
        Initializes the FBSDESolver.

        Args:
        - pde : the partial differential equation to solve
        - layer_sizes (list[int]): list of sizes of hidden layers
        - learning_rate (float): learning rate for optimization
        - solving_method (str): method to solve the PDE ('Monte-Carlo', 'Deep-Time-SetTransformer', 'Basic')
        """
        self.pde = pde
        self.layer_size = layer_sizes
        self.learning_rate = learning_rate
        self.device = device



        self.solver = FBSNNNetwork(self.layer_size)


        self.optimizer = torch.optim.Adam(self.solver.parameters(), lr=self.learning_rate,betas=(0.99, 0.999))
    def compute_loss(self,
                     y,
                     dw,
                     t,
                     init
    ):
      batch_size = y.size()[0]
      y_terminal = torch.ones((batch_size,)).to(self.device) * init.to(self.device)

      y_terminal = y_terminal.to(self.device)
      L1 = 0.0
      loss = 0
      for z in range(self.pde.num_time_interval):
          S0 = y[:, :, z]
          S1 = y[:, :, z+1]

          t0 = t * self.pde.delta_t * z
          t1  = t * self.pde.delta_t * (z+1)

          y_0 = torch.squeeze(self.solver(torch.cat((t0, S0),dim = 1) ))
          y_1 =  torch.squeeze(self.solver(torch.cat((t1, S1),dim = 1) ))

          x = torch.tensor(S0,requires_grad=True)
          out1 = torch.autograd.grad(outputs=torch.sum(self.solver(torch.cat((t0, x),dim = 1) )), inputs=x ,  create_graph=True)[0]
          Z = out1 * self.pde.sigma_matrix(S0)

          #out1 = torch.autograd.grad(self.solver(torch.cat((t0, x),dim = 1) ), x, torch.ones_like(self.solver(torch.cat((t0, x),dim = 1) )), create_graph=True, retain_graph=True)
          #print(torch.autograd.grad(outputs=torch.sum(out1[0]), inputs=self.solver.parameters(),create_graph=True, allow_unused=True)[0] )

          samp = dw[:, :, z].to(self.device)

          nsim = S0.size()[0]
          interest_rate = torch.squeeze(self.pde.r_u(t0, S0, y_0, Z).to(self.device))
          hz = self.pde.h_z(t0, S0, y_0, Z)
          y_1_hat = y_0 * (1 + interest_rate * self.pde.delta_t) + torch.squeeze(
              self.pde.h_z(t0, S0, y_0, Z).to(self.device)) * self.pde.delta_t + torch.squeeze(
              torch.bmm(torch.unsqueeze(Z, 1), torch.unsqueeze(samp, 2)))
          loss += torch.mean(torch.square(torch.squeeze(y_1_hat) - torch.squeeze(y_1)))


      S0 = y[:, :, self.pde.num_time_interval]
      t0 = t * self.pde.delta_t * self.pde.num_time_interval

      payoff = self.pde.terminal(t0, S0)
      y_terminal = torch.squeeze(self.solver(torch.cat((t0, S0),dim = 1) ))



      loss += torch.mean(torch.square(torch.squeeze(payoff) - torch.squeeze(y_terminal)))
      #x = torch.tensor(S0,requires_grad=True)
      #out1 = torch.autograd.grad(outputs=torch.sum( torch.squeeze(self.solver(torch.cat((t0, x),dim = 1) ))), inputs=x ,  create_graph=True)[0]

      #out1_exact = torch.autograd.grad(outputs=torch.sum(torch.squeeze(self.pde.terminal(t0, x))), inputs=x )[0]
      #loss += torch.mean(torch.square(torch.squeeze(out1) - torch.squeeze(out1_exact))) * 100
      return loss

    def solve(
            self,
            num_iterations,
            batch_size,
            init,
            sample_size=None
    ):
        """
        Solves the PDE.

        Args:
        - num_iterations (int): number of iterations for optimization
        - batch_size (int): batch size for training
        - init (torch.Tensor): initial value
        - device (torch.device): device to perform calculations on ('cpu', 'cuda')
        - sample_size (int, optional): sample size for Monte-Carlo method

        Returns:
        - list[torch.Tensor]: list of losses during optimization
        - list[torch.Tensor]: list of initial values during optimization
        """
        losses = []
        inits = []
        self.solver.to(self.device)

        torch.seed()
        self.solver.train()
        dw_val, y_val = self.pde.sample(128)
        y_val = torch.tensor(y_val)
        y_val = y_val.to(torch.float32).to(self.device)
        dw_val = torch.tensor(dw_val).to(torch.float32)
        dw_val = dw_val.to(self.device)
        t_val = torch.ones((128, 1)).to(self.device)

        for j in range(num_iterations):
            print(j + 1)

            dw, y = self.pde.sample(batch_size)
            y = torch.tensor(y)
            y = y.to(torch.float32)
            dw = torch.tensor(dw).to(torch.float32)
            dw = dw.to(self.device)
            t = torch.ones((batch_size, 1)).to(self.device)

            self.optimizer.zero_grad()

            y = y.to(self.device)

            loss = self.compute_loss(y,dw,t,init)
            loss.backward()

            self.optimizer.step()



            loss = self.compute_loss(y_val,dw_val,t_val,init)
            S0 = y_val[:, :, 0]
            t0 = t_val * self.pde.delta_t  *0

            init = torch.squeeze(self.solver(torch.cat((t0, S0),dim = 1) ))[0]
            print(init)
            inits.append(init.detach().cpu().numpy())




            losses.append(loss.detach().cpu().numpy())

            print(loss)


        return losses,inits


from collections import Counter
from math import log2
from typing import Dict, Tuple


def longest_match_length(
    message: str,
    i: int,
    n: int
) -> Tuple[int, str]:
    """
    Calculate the length of the longest match.

    :param message: Input encoded message
    :type message: str
    :param i: Index value
    :type i: int
    :param n: Length parameter
    :type n: int
    :return: Tuple containing matched length and substring
    :rtype: tuple
    """
    longest_match = ""
    for l in range(1, n + 1):
        pattern = message[i:i + l + 1]
        for j in range(i - n + 1, i + 1):
            candidate = message[j:j + l + 1]
            if pattern == candidate:
                longest_match = pattern
                break

    return len(longest_match) + 1, longest_match


def kontoyiannis_entropy(
    message: str,
    window: int = None
) -> float:
    """
    Calculate Kontoyiannis Entropy.

    :param message: Input encoded message
    :type message: str
    :param window: Length of expanding window, default is None
    :type window: int or None
    :return: Calculated Kontoyiannis Entropy
    :rtype: float
    """
    output = {"num": 0, "sum": 0, "sub_string": []}
    message_length = len(message)

    if window is None:
        points = range(2, message_length // 2 + 2)
    else:
        window = min(window, message_length // 2)
        points = range(window + 1, message_length - window + 2)

    for i in points:
        n = i if window is None else window
        l, sub_string = longest_match_length(message, i, n)
        output["sum"] += log2(n) / l
        output["sub_string"].append(sub_string)
        output["num"] += 1

    output["h"] = output["sum"] / output["num"]
    output["r"] = 1 - output["h"] / log2(message_length)

    return output["h"]
